#!/usr/bin/env python
#
# This file is part of Cynthion.
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Cynthion 'run' command. """

import logging, os, subprocess, sys, tempfile

from .util  import find_cynthion_asset, find_cynthion_bitstream
from .util  import flash_soc_firmware, run_bitstream


def cynthion_run(device, args):
    if args.bitstream is not None:
        run_bitstream(device, args.bitstream)

    elif args.target == "analyzer":
        run_bitstream(device, find_cynthion_bitstream(device, f"analyzer.bit"))

    elif args.target == "facedancer":
        flash_soc_firmware(device, find_cynthion_asset("moondancer.bin"))
        run_bitstream(device, find_cynthion_bitstream(device, f"facedancer.bit"))

    elif args.target == "selftest":
        _run_selftest(device, args)

    else:
        logging.error(f"unknown target: {args.target}")
        return


def _run_selftest(device, args):
    from cynthion.selftest.gateware import SelftestDevice
    from cynthion.selftest.host import StandaloneTester

    run_bitstream(device, find_cynthion_bitstream(device, f"selftest.bit"))
    device.close()

    selftest_device = SelftestDevice()

    tester = StandaloneTester(selftest_device)
    tester.run_tests()
    print()



# This file is part of Cynthion.
#
# Copyright (c) 2020-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import print_function
# Alias objects to make them easier to import.

# mildly evil hack to vendor in amaranth_stdio for the benefit of
# apollo_fpga.gateware.advertiser.ApolloAdvertiser
try:
    try:
        import amaranth_stdio
    except:
        import sys
        from luna_soc.gateware.vendor import amaranth_stdio
        sys.modules["amaranth_stdio"] = amaranth_stdio
except:
    pass

from .cynthion import Cynthion
from .cynthion import CynthionSingleton
from .cynthion import CynthionBoard

from . import gateware
from . import shared

import importlib.metadata
__version__ = importlib.metadata.version(__package__)


Cynthion = Cynthion  # pyflakes


class _CynthionSingletonWrapper(object):

    """
    Convenience function that acts like CynthionSingleton, but also allows Magic:
    accessing a property on this object will act as though that property had been
    accessed on a result of a CynthionSingleton() call.

    That's heckin' unreadable, so in short-- accessing a property on a relevant object
    will attempt to 1) call the property on the sanest existing Cynthion object; or
    2) create a new Cynthion object, if necessary.
    """

    def __init__(self, serial=None):
        self.serial = serial

    def __getitem__(self, serial):
        return _CynthionSingletonWrapper(serial)

    def __getattr__(self, name):
        return getattr(CynthionSingleton(self.serial), name)

    def __call__(self, serial=None):
        return CynthionSingleton(serial)

    def __dir__(self):
        return dir(CynthionSingleton(self.serial))

CynthionSingleton = _CynthionSingletonWrapper()


# TODO deprecate in favor of:
#
#   from importlib.resources import files
#   assets_directory = files("cynthion").joinpath("assets")
#
def cynthion_assets_directory():
    """ Provide a quick function that helps us get at our assets directory. """
    import os

    # Find the path to the module, and then find its assets folder.
    module_path = os.path.dirname(__file__)
    return os.path.join(module_path, 'assets')


def find_cynthion_asset(filename):
    """ Returns the path to a given Cynthion asset, if it exists, or None if the Cynthion asset isn't provided."""
    import os

    asset_path = os.path.join(cynthion_assets_directory(), filename)

    if os.path.isfile(asset_path):
        return asset_path
    else:
        return None



#
# This file is part of Cynthion
#


from enum import IntEnum
from warnings import warn

from ..interface import CynthionInterface


# TODOs:
#  - XXX: Overhaul the GPIO(Collection) class to be more efficient
#  - More cleanup to use the GPIOPin model.
#  - Support ranges of pins from the same port (GPIOPort objects?)
#  - Implement a release function so if e.g. an I2C device is no longer in use
#    it releases its pins back to the GPIO pool.

class Directions(IntEnum):
    IN = 0
    OUT = 1


# Legacy convenience constants.
DIRECTION_IN  = Directions.IN
DIRECTION_OUT = Directions.OUT



class GPIOProvider(CynthionInterface):
    """ Base class for an object that provides access to GPIO pins. """

    # For convenience.
    DIRECTION_IN  = Directions.IN
    DIRECTION_OUT = Directions.OUT

    # If the subclass has a fixed set of pins, it can override this mapping to
    # specify the fixed pin names to be automatically registered.
    FIXED_GPIO_PINS = {}

    # If the subclass doesn't want to allow external sources to register GPIO pins
    ALLOW_EXTERNAL_REGISTRATION = True


    def __init__(self, name_mappings=None):
        """ Sets up the basic fields for a GPIOProvider.

        Parameters:
            name_mappings -- Allows callers to rename the local / fixed GPIO pin names.
                Optional; accepts a dictionary mapping their fixed names to their new names, or
                to None to remove the relevant pin from the list of available pins.

                This allows instantiators to give a given GPIO collection more specific names, or
                to hide them from general API display/usage.
        """

        if name_mappings is None:
            name_mappings = {}

        # Set up our basic tracking parameters, which track which GPIO pins
        # are available and in use.
        self.pin_mappings   = {}
        self.active_gpio    = {}
        self.available_pins = []

        # Add all of our fixed pins as acceptable GPIO.
        for name, line in self.FIXED_GPIO_PINS.items():

            # If we've been asked to rename the given pin, register it under
            # the new name, rather than under the provided name,
            if name in name_mappings:
                name = name_mappings[name]

            # If our name field winds up remapping to 'None', the instantiator
            # is trying to hide the relevant pin. Skip registering it.
            if name is None:
                continue

            # Register each fixed GPIO.
            self.__register_gpio(name, line)


    def register_gpio(self, name, line, used=False):
        """
        Registers a GPIO pin for later use. Usually only used in board setup.

        Args:
            name -- The name for the GPIO, usually expressed as a position on
                a Cynthion header.
            line -- An abstract argument passed to subclass methods that serves
                to identify the pin. Subclasses often use this to store e.g. port and pin
                numbers.
        """

        # If this class doesn't allow pin registration, raise an error.
        if not self.ALLOW_EXTERNAL_REGISTRATION:
            raise NotImplementedError("This GPIO collection does not allow registration of new pins.")

        # Otherwise, delegate to our internal registration method.
        self.__register_gpio(name, line, used)


    def __register_gpio(self, name, line, used=False):
        """
        Registers a GPIO pin for later use. Usually only used in board setup.

        Args:
            name -- The name for the GPIO, usually expressed as a position on
                a Cynthion header.
            line -- An abstract argument passed to subclass methods that serves
                to identify the pin. Subclasses often use this to store e.g. port and pin
                numbers.
        """

        # Store the full name in our pin mappings.
        self.pin_mappings[name] = line

        if not used:
            self.mark_pin_as_unused(name)


    def mark_pin_as_used(self, name):
        """ Marks a pin as used by another peripheral. """

        if name not in self.pin_mappings:
            raise ValueError("Unknown GPIO pin {}".format(name))

        self.available_pins.remove(name)


    def mark_pin_as_unused(self, name):
        """ Mark a pin as no longer used by another peripheral. """

        if name not in self.pin_mappings:
            raise ValueError("Unknown GPIO pin {}".format(name))

        if name not in self.available_pins:
            self.available_pins.append(name)


    def get_available_pins(self, include_active=True):
        """ Returns a list of available GPIO names. """
        available = self.available_pins[:]
        available.extend(self.active_gpio.keys())

        return available


    def get_pin(self, name, unique=False):
        """
        Returns a GPIOPin object by which a given pin can be controlled.

        Args:
            name -- The GPIO name to be used.
            unique -- True if this should fail if a GPIO object for this pin
                already exists.
        """

        # If we already have an active GPIO pin for the relevant name, return it.
        if name in self.active_gpio and not unique:
            return self.active_gpio[name]

        # If the pin's available for GPIO use, grab it.
        if name in self.available_pins:
            port = self.pin_mappings[name]

            self.active_gpio[name] = GPIOPin(self, name, port)
            self.mark_pin_as_used(name)

            return self.active_gpio[name]

        # If we couldn't create the GPIO pin, fail out.
        raise ValueError("No available GPIO pin {}".format(name))



    def get_port(self, *pin_names):
        """ Creates a GPIOPort object that can set multiple pins to a binary value.

        Arguments are a list of pin names to conglomerate into a port, MSB first. This may result in a GPIOPort
        object, or in a derivative class such as a VirtualGPIOPort, depending on the pin locations.
        """

        pins = []

        # Convert each of the header pin names to a GPIOPin object.
        for name in pin_names:
            pins.append(self.get_pin(name))


        # FIXME: apply an optimization for when each pin is on the same logical port:
        return VirtualGPIOPort(pins)


    def release_pin(self, gpio_pin):
        """
        Releases a GPIO pin back to the system for re-use, potentially
        not as a GPIO.
        """

        if gpio_pin.name not in self.active_gpio:
            raise ValueError("Trying to release a pin we don't own!")

        # Mark the pin as an input, placing it into High-Z mode.
        # TODO: Disable any pull-ups present on the pin.
        gpio_pin.set_direction(DIRECTION_IN)

        # Remove the GPIO pin from our active array, and add it back to the
        # available pool.
        del self.active_gpio[gpio_pin.name]
        self.mark_pin_as_unused(gpio_pin.name)


    def set_up_pin(self, line, direction, initial_value=False):
        """
        Configure a GPIO line for use as an input or output.  This must be
        called before the line can be used by other functions.

        Parameters:
            line      -- A unique identifier for the given pin that has meaning to the subclass.
            direction -- Directions.IN (input) or Directions.OUT (output)
        """
        pass


    def set_pin_state(self, line, state):
        """
        Set the state of an output line.  The line must have previously been
        configured as an output using setup().

        Parameters:
            line  -- A unique identifier for the given pin that has meaning to the subclass.
            state -- True sets line high, False sets line low
        """
        pass


    def read_pin_state(self, line):
        """
        Get the state of an input line.  The line must have previously been
        configured as an input using setup().

        Args:
            line  -- A unique identifier for the given pin that has meaning to the subclass.

        Return:
            bool -- True if line is high, False if line is low
        """
        pass


    def get_pin_direction(self, line):
        """
        Gets the direction of a GPIO pin.

        Args:
            line  -- A unique identifier for the given pin that has meaning to the subclass.

        Return:
            bool -- True if line is an output, False if line is an input
        """
        pass


    def get_pin_port(self, line):
        """ Returns the 'port number' for a given GPIO pin.

        For providers for which 'port' isn't a valid semantic concept, this should return
        the same identifier for every pin that can be logically written in a single operation.
        """
        pass


    def get_pin_identifier(self, line):
        """ Returns the 'pin number' for a given GPIO pin.

        This number is typically the 'bit number' in a larger, organized port. For providers
        in which this isn't a valid semantic concept, any convenient semantic identifier (or None)
        is acceptable.
        """
        pass


class GPIO(GPIOProvider):
    """ Work with the GPIO directly present on the Cynthion board. """


    def __init__(self, board):
        """
        Args:
            board -- Cynthion board whose GPIO lines are to be controlled
        """

        # Set up our basic fields...
        super(GPIO, self).__init__()

        # ... and store information about the our low-level connection.
        self.board = board
        self.api   = self.board.apis.gpio

        # TODO: provide functionality to restore GPIO state on reconnect?


    def set_up_pin(self, line, direction, initial_value=False):
        """
        Configure a GPIO line for use as an input or output.  This must be
        called before the line can be used by other functions.

        Args:
            line -- (port, pin); typically a tuple from J1, J2, J7 below
            direction -- Directions.IN (input) or Directions.OUT (output)

        TODO: allow pull-up/pull-down resistors to be configured for inputs
        """
        self.api.set_up_pin(line[0], line[1], direction, initial_value)


    def set_pin_state(self, line, state):
        """
        Set the state of an output line.  The line must have previously been
        configured as an output using setup().

        Args:
            line -- (port, pin); typically a tuple from J1, J2, J7 below
            state -- True sets line high, False sets line low
        """

        # TODO: validate GPIO direction?

        single_write = (line[0], line[1], state,)
        self.api.write_pins(single_write)



    def read_pin_state(self, line):
        """
        Get the state of an input line.  The line must have previously been
        configured as an input using setup().

        Args:
            line -- (port, pin); typically a tuple from J1, J2, J7 below

        Return:
            bool -- True if line is high, False if line is low
        """
        values = self.api.read_pins(line)
        return values[0]


    def get_pin_direction(self, line):
        """
        Gets the direction of a GPIO pin.

        Args:
            line -- (port, pin); typically a tuple from J1, J2, J7 below

        Return:
            bool -- True if line is an output, False if line is an input
        """
        directions = self.api.get_pin_directions(line)
        return directions[0]

    def get_pin_port(self, line):
        """ Returns the 'port number' for a given GPIO pin."""
        return line[0]

    def get_pin_identifier(self, line):
        """ Returns the 'pin number' for a given GPIO pin. """
        return line[1]



    #
    # Deprecated methods.
    #
    def output(self, line, state):
        warn("GPIO.output is deprecated; prefer set_pin_state.", DeprecationWarning)
        self.set_pin_state(line, state)

    def input(self, line):
        warn("GPIO.input is deprecated; prefer read_pin_state.", DeprecationWarning)
        return self.read_pin_state(line)

    def setup(self, line, direction):
        warn("GPIO.setup is deprecated; prefer set_up_pin.", DeprecationWarning)
        self.set_up_pin(line, direction)




class GPIOPin(object):
    """
    Class representing a single GPIO pin.
    """

    def __init__(self, gpio_provider, name, line):
        """
        Creates a new object representing a GPIO Pin. Usually instantiated via
        a GPIO object.

        Args:
            gpio_provider -- The GPIO object to which this pin belongs.
            name -- The name of the given pin. Should match a name registered
                in its GPIO collection.
            line -- The pin's 'line' information, as defined by the object that created
                this GPIO pin. This variable has semantic meaning to the GPIO collection;
                but doesn't have any semantic meaning to this class.

        """

        self.name    = name
        self._parent = gpio_provider
        self._line   = line

        # For convenience:
        self.DIRECTION_IN  = Directions.IN
        self.DIRECTION_OUT = Directions.OUT

        # Set up the pin for use. Idempotent.
        self._parent.set_up_pin(self._line, self.get_direction(), self.read())


    def set_direction(self, direction, initial_value=False):
        """
        Sets the GPIO pin to use a given direction.
        """
        self._parent.set_up_pin(self._line, direction, initial_value)


    def get_direction(self):
        """ Returns the pin's direction; will be either Directions.IN or Directions.OUT """
        return self._parent.get_pin_direction(self._line)


    def is_input(self):
        """ Convenience function that returns True iff the pin is configured as an input. """
        return (self.get_direction() == self.DIRECTION_IN)


    def is_output(self):
        """ Convenience function that returns True iff the pin is configured as an output. """
        return (self.get_direction() == self.DIRECTION_OUT)


    def read(self, high_value=True, low_value=False, check_pin_direction=False, set_pin_direction=False):
        """ Convenience alias for get_state."""
        return self.get_state(high_value, low_value, check_pin_direction, set_pin_direction)


    def input(self, high_value=True, low_value=False):
        """ Convenience function that sets the pin to an input and reads its value. """
        return self.read(set_pin_direction=True, high_value=high_value, low_value=low_value)


    def get_state(self, high_value=True, low_value=False, check_pin_direction=False, set_pin_direction=False):
        """ Returns the value of a GPIO pin. """

        # If we're setting the pin direction while we're getting the state, set it.
        if set_pin_direction:
            self.set_direction(self.DIRECTION_IN)

        # Otherwise, enforce the direction, if desired.
        elif check_pin_direction and not self.is_input():
            raise ValueError("Trying to read from a non-input pin {}! Set up the pin first with set_direction.".format(self.name))

        # Finally, read the pin's state.
        raw = self._parent.read_pin_state(self._line)
        return high_value if raw else low_value


    def write(self, high, check_direction=False):
        """ Convenience alias for set_state."""
        self.set_state(high, check_direction)


    def set_state(self, high, check_direction=True):
        """ Write a given value to the GPIO port.

        Args:
            high -- True iff the pin should be set to high; the pin will be set
                to low otherwise.
        """

        if check_direction and not self.is_output():
            raise ValueError("Trying to write to a non-output pin {}! Set up the pin first with set_direction.".format(self.name))

        self._parent.set_pin_state(self._line, high)


    def high(self):
        """ Convenience function that sets the given GPIO pin to both output mode and high, at once. """

        # Note that we can't rely on initial_direction to set the actual port value; as some
        # GPIOProviders may not support that.
        self.set_direction(self.DIRECTION_OUT, True)
        self.write(True)


    def low(self):
        """ Convenience function that sets the given GPIO pin to both output mode and low, at once. """

        # Note that we can't rely on initial_direction to set the actual port value; as some
        # GPIOProviders may not support that.
        self.set_direction(self.DIRECTION_OUT, False)
        self.write(False)


    def get_port(self):
        """ Returns device's port number, if possible. """
        return self._parent.get_pin_port(self._line)


    def get_pin(self):
        """ Returns pin's pin number within its port, if possible. """
        return self._parent.get_pin_identifier(self._line)


    # TODO: Toggle-- we have the hardware for this :)

    # TODO: handle pulldowns/pull-ups, etc.



class VirtualGPIOPort(object):
    """ An object that represents a "virtually contiguous" group of GPIO pins. """


    def __init__(self, *pin_arguments):
        """ Creates a virtual GPIO Port from GPIOPin-compatible objects.

            pins -- A list of pins to be coalesced into a virtual port;
                with the MSB first. For convenience, pins (or lists) can
                also be provided as variadic arguments. Pins should already
                have their directions / resistors set.
        """

        pins = []

        # Take each of our passed in pins/objects, and add them to our ordered list.
        for pin in pin_arguments:
            if isinstance(pin, list):
                pins.extend(pin)
            else:
                pins.append(pin)

        # Reverse the order of our list, so element 0 corresponds to bit zero.
        self.pins = pins[::-1]


    def set_direction(self, word, initial_value=0):
        """ Sets the direction of each individual pin.

        Parameters:
            word -- A number whose bits contain 1s for each bit that should be an output,
                and zeroes for each bit that should be an input.
        """

        for bit, pin in enumerate(self.pins):
            direction = DIRECTION_OUT if (word & (1 << bit)) else DIRECTION_IN
            initial_value = bool(initial_value & (1 << bit))

            pin.set_direction(direction, initial_value=initial_value)


    def all_output(self, initial_value=False):
        """ Sets all of the pins in this port to output mode.

        Parameters:
            initial_value -- Optional; the start value to apply to each pin.
        """

        for pin in self.pins:
            pin.set_direction(DIRECTION_OUT, initial_value)


    def all_input(self):
        """ Sets all of the pins in this port to output mode. """

        for pin in self.pins:
            pin.set_direction(DIRECTION_IN)



    def read(self):
        """ Returns the integer value of the relevant port. """

        value = 0

        # Iterate over each of the contained pins, and add it to our value.
        for bit, pin in enumerate(self.pins):

            # If this pin reads as true, add it to our composite.
            if pin.read():
                value |= (1 << bit)

        return value


    def write(self, value):
        """ Writes a given integer value to the port. """

        for bit, pin in enumerate(self.pins):
            new_value = bool(value & (1 << bit))
            pin.write(new_value)



#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

import os

from amaranth.build import *
from ..vendor.amaranth_boards.resources import *

from .core import CynthionPlatform

__all__ = ["CynthionPlatformRev0D3"]

#
# Note that r0.3 have D+/D- swapped to avoid having to cross D+/D- in routing.
#
# This is supported by a PHY feature that allows you to swap pins 13 + 14.
#

class CynthionPlatformRev0D3(CynthionPlatform):
    """ Board description for the pre-release r0.3 revision of Cynthion. """

    name        = "Cynthion r0.3"
    version     = (0, 3)
    device      = "LFE5U-12F"
    package     = "BG256"
    speed       = os.getenv("ECP5_SPEED_GRADE", "8")

    # By default, assume we'll be connecting via our target PHY.
    default_usb_connection = "host_phy"

    #
    # Preferred DRAM bus I/O (de)-skewing constants.
    #
    ram_timings = dict(
        clock_skew = 16
    )

    #
    # I/O resources.
    #
    resources   = [

        # Primary, discrete 60MHz oscillator.
        Resource("clk_60MHz", 0, Pins("A8", dir="i"),
            Clock(60e6), Attrs(IO_TYPE="LVCMOS33")),

        # Connection to our SPI flash; can be used to work with the flash
        # from e.g. a bootloader.
        Resource("spi_flash", 0,

            # SCK is on pin 9; but doesn't have a traditional I/O buffer.
            # Instead, we'll need to drive a clock into a USRMCLK instance.
            # See interfaces/flash.py for more information.
            Subsignal("sdi",  Pins("T8",  dir="o")),
            Subsignal("sdo",  Pins("T7",  dir="i")),
            Subsignal("cs",   PinsN("N8", dir="o")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # Connection to our SPI flash but using quad mode (QSPI)
        Resource("qspi_flash", 0,
            # SCK is on pin 9; but doesn't have a traditional I/O buffer.
            # Instead, we'll need to drive a clock into a USRMCLK instance.
            # See interfaces/flash.py for more information.
            Subsignal("dq",  Pins("T8 T7 M7 N7",  dir="io")),
            Subsignal("cs",  PinsN("N8", dir="o")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # Note: UART pins R14 and T14 are connected to JTAG pins R11 (TDI)
        # and T11 (TMS) respectively, so the microcontroller can use either
        # function but not both simultaneously.

        # UART connected to the debug controller; can be routed to a host via CDC-ACM.
        Resource("uart", 0,
            Subsignal("rx",  Pins("R14",  dir="i")),
            Subsignal("tx",  Pins("T14",  dir="oe"), Attrs(PULLMODE="UP")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # SPI bus connected to test points for simple register exchanges.
        # The FPGA acts as peripheral, not controller.
        Resource("debug_spi", 0,
            Subsignal("sck",  Pins( "R13", dir="i")),
            Subsignal("sdi",  Pins( "P13", dir="i")),
            Subsignal("sdo",  Pins( "P11", dir="o")),
            Subsignal("cs",   PinsN("T13", dir="i")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # FPGA-connected LEDs numbered 5-0.
        *LEDResources(pins="P14 P16 P15 R16 R15 T15", attrs=Attrs(IO_TYPE="LVCMOS33"), invert=True),

        # USB PHYs
        ULPIResource("target_phy", 0,
            data="E16 F14 F16 F15 G16 G15 H15 J16", clk="C15", clk_dir='o',
            dir="D16", nxt="E15", stp="D14", rst="C16", rst_invert=True,
            attrs=Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
        ULPIResource("sideband_phy", 0,
            data="R1 P3 P1 P2 N1 M2 M1 L2", clk="P4", clk_dir='o',
            dir="T2", nxt="R2", stp="R3", rst="T3", rst_invert=True,
            attrs=Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
        ULPIResource("host_phy", 0,
            data="F1 F2 E1 E2 D1 E3 C1 C2", clk="J1", clk_dir='o',
            dir="G1", nxt="G2", stp="H2", rst="J2", rst_invert=True,
            attrs=Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),

        # Target port power switching.
        #
        # power_c_port does the reverse of pass_through_vbus, passing power
        # from the Type-A port to the Type-C port. It is intended to be used in
        # conjunction with power_a_port, supplying VBUS to both target ports.

        Resource("power_a_port",         0, Pins("C14", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("power_c_port",         0, Pins("F13", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("pass_through_vbus",    0, Pins("B16", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("target_c_to_a_fault",  0, Pins("F12", dir="i"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("target_a_to_c_fault",  0, Pins("E14", dir="i"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("target_5v_to_a_fault", 0, Pins("B15", dir="i"), Attrs(IO_TYPE="LVCMOS33")),

        # HyperRAM (3V3 domain).
        Resource("ram", 0,
            # Note: our clock uses the pseudo-differential I/O present on the top tiles.
            # This requires a recent version of trellis+nextpnr. If your build complains
            # that LVCMOS33D is an invalid I/O type, you'll need to upgrade.
            Subsignal("clk",   DiffPairs("B14", "A15", dir="o"), Attrs(IO_TYPE="LVCMOS33D")),
            Subsignal("dq",    Pins("A11 B10 B12 A12 B11 A10 B9 A9", dir="io")),
            Subsignal("rwds",  Pins( "A13", dir="io")),
            Subsignal("cs",    PinsN("A14", dir="o")),
            Subsignal("reset", PinsN("B13", dir="o")),
            Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")
        ),

        # User I/O connections (SMA connectors).
        Resource("user_io", 0, Pins("C3", dir="io"), Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
        Resource("user_io", 1, Pins("D3", dir="io"), Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),

        # Convenience references.
        Resource("user_pmod", 0, Pins("1 2 3 4 7 8 9 10", conn=("pmod", 0), dir="io"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("user_pmod", 1, Pins("1 2 3 4 7 8 9 10", conn=("pmod", 1), dir="io"), Attrs(IO_TYPE="LVCMOS33")),
    ]

    connectors = [
        Connector("pmod", 0, "A3 A4 A5 A6 - - C6 B6 C7 B7 - -"), # Pmod A
        Connector("pmod", 1, "M5 N5 M4 N3 - - L4 L5 K4 K5 - -"), # Pmod B
    ]



#
# This file is part of Cynthion.
#

"""
The values in this module are generated at runtime from the
corresponding TOML files located in the `cynthion.git/host/cynthion/shared/`
directory.
"""

import glob
import tomli

from collections import namedtuple
try:
    # <= 3.8
    from importlib_resources import files
except:
    # >= 3.9
    from importlib.resources import files
from os import path
from pathlib import Path


SHARED_TOML_PATH = Path(files("cynthion").joinpath("shared"))


def generate_module_values():
    for toml in SHARED_TOML_PATH.glob("*.toml"):
        with toml.open("rb") as f:
            globals()[toml.stem] = _dict_to_namedtuple(tomli.load(f))


def _dict_to_namedtuple(data, typename="_"):
    return namedtuple(typename, data.keys())(
        *(_dict_to_namedtuple(v, typename + '_' + k) if isinstance(v, dict) else v for k, v in data.items())
    )

generate_module_values()



#
# This file is part of Cynthion.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from amaranth               import Cat, Elaboratable, Module, Signal, Record
from luna.gateware.platform import NullPin
from luna_soc.gateware.csr  import Peripheral

from apollo_fpga.gateware   import ApolloAdvertiser

class ApolloAdvertiserPeripheral(Peripheral, Elaboratable):
    """ Controller peripheral for ApolloAdvertiser

    CSR registers
    -------------
    enable : read/write
        Set this bit to '1' to start ApolloAdvertiser and disconnect the Cynthion USB control port from Apollo.
    """

    def __init__(self, clk_freq_hz=None, pad=None):
        super().__init__()

        self.advertiser = ApolloAdvertiser(pad=pad, clk_freq_hz=clk_freq_hz)

        #
        # Registers
        #

        regs         = self.csr_bank()
        self._enable = regs.csr(1, "rw", desc="""
            Set this bit to '1' to start ApolloAdvertiser and disconnect the Cynthion USB control port from Apollo.
        """)

        # wishbone connection
        self._bridge = self.bridge(data_width=32, granularity=8, alignment=2)
        self.bus     = self._bridge.bus


    def elaborate(self, platform):
        m = Module()
        m.submodules.bridge = self._bridge

        # advertiser stop signal is enabled by default
        stop = Signal(reset=1)

        # update advertiser stop signal on register write
        with m.If(self._enable.w_stb):
            m.d.sync += stop.eq(~self._enable.w_data)

        # create advertiser
        m.submodules.advertiser = self.advertiser
        m.d.comb += self.advertiser.stop.eq(stop)

        return m



#!/usr/bin/env python3
#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from amaranth import Signal, Elaboratable, Module, Cat, ClockDomain, ClockSignal, ResetSignal
from amaranth.lib.cdc import FFSynchronizer

from luna.gateware.utils.cdc          import synchronize
from luna.gateware.architecture.car   import LunaECP5DomainGenerator
from luna.gateware.interface.jtag     import JTAGRegisterInterface
from luna.gateware.interface.ulpi     import ULPIRegisterWindow
from luna.gateware.interface.psram    import HyperRAMPHY, HyperRAMInterface

from .registers import *


CLOCK_FREQUENCIES = {
    "fast": 60,
    "sync": 60,
    "usb":  60
}


class SelftestDevice(Elaboratable):

    def elaborate(self, platform):
        m = Module()

        # Generate our clock domains.
        clocking = LunaECP5DomainGenerator(clock_frequencies=CLOCK_FREQUENCIES)
        m.submodules.clocking = clocking

        registers = JTAGRegisterInterface(default_read_value=0xDEADBEEF)
        m.submodules.registers = registers

        # Simple applet ID register.
        registers.add_read_only_register(REGISTER_ID, read=0x54455354)

        # LED test register.
        led_reg = registers.add_register(REGISTER_LEDS, size=6, name="leds", reset=0b111111)
        led_out   = Cat([platform.request("led", i, dir="o").o for i in range(0, 6)])
        m.d.comb += led_out.eq(led_reg)

        #
        # ULPI PHY windows
        #
        self.add_ulpi_registers(m, platform,
            ulpi_bus="target_phy",
            register_base=REGISTER_TARGET_ADDR
        )
        self.add_ulpi_registers(m, platform,
            ulpi_bus="aux_phy" if platform.version >= (0, 6) else "host_phy",
            register_base=REGISTER_AUX_ADDR
        )
        self.add_ulpi_registers(m, platform,
            ulpi_bus="control_phy" if platform.version >= (0, 6) else "sideband_phy",
            register_base=REGISTER_CONTROL_ADDR
        )


        #
        # HyperRAM test connections.
        #
        ram_bus = platform.request('ram')
        psram_phy = HyperRAMPHY(bus=ram_bus)
        psram = HyperRAMInterface(phy=psram_phy.phy)
        m.submodules += [psram_phy, psram]

        psram_address_changed = Signal()
        psram_address = registers.add_register(REGISTER_RAM_REG_ADDR, write_strobe=psram_address_changed)

        # Store last read word from HyperRAM.
        psram_read_data = Signal.like(psram.read_data)
        with m.If(psram.read_ready):
            m.d.sync += psram_read_data.eq(psram.read_data)
        registers.add_sfr(REGISTER_RAM_VALUE, read=psram_read_data)

        # Hook up our PSRAM.
        m.d.comb += [
            ram_bus.reset.o        .eq(0),
            psram.single_page      .eq(0),
            psram.perform_write    .eq(0),
            psram.register_space   .eq(1),
            psram.final_word       .eq(1),
            psram.start_transfer   .eq(psram_address_changed),
            psram.address          .eq(psram_address),
        ]

        return m


    def add_ulpi_registers(self, m, platform, *, ulpi_bus, register_base):
        """ Adds a set of ULPI registers to the active design. """

        target_ulpi      = platform.request(ulpi_bus)

        ulpi_reg_window  = ULPIRegisterWindow()
        m.submodules  += ulpi_reg_window

        m.d.comb += [
            ulpi_reg_window.ulpi_data_in  .eq(target_ulpi.data.i),
            ulpi_reg_window.ulpi_dir      .eq(target_ulpi.dir.i),
            ulpi_reg_window.ulpi_next     .eq(target_ulpi.nxt.i),

            target_ulpi.clk.o    .eq(ClockSignal("usb")),
            target_ulpi.rst.o    .eq(ResetSignal("usb")),
            target_ulpi.stp.o    .eq(ulpi_reg_window.ulpi_stop),
            target_ulpi.data.o   .eq(ulpi_reg_window.ulpi_data_out),
            target_ulpi.data.oe  .eq(~target_ulpi.dir.i)
        ]

        register_address_change  = Signal()
        register_value_change    = Signal()

        # ULPI register address.
        registers = m.submodules.registers
        registers.add_register(register_base + 0,
            write_strobe=register_address_change,
            value_signal=ulpi_reg_window.address,
            size=6
        )
        m.submodules.clocking.stretch_sync_strobe_to_usb(m,
            strobe=register_address_change,
            output=ulpi_reg_window.read_request,
        )

        # ULPI register value.
        registers.add_sfr(register_base + 1,
            read=ulpi_reg_window.read_data,
            write_signal=ulpi_reg_window.write_data,
            write_strobe=register_value_change
        )
        m.submodules.clocking.stretch_sync_strobe_to_usb(m,
            strobe=register_value_change,
            output=ulpi_reg_window.write_request
        )



#
# This file is part of Cynthion
#

import time

class CynthionInterface(object):
    """
    Generic base class for Cynthion peripherals.
    """


    def __init__(self, device):
        """ Default peripheral initializer -- just stores a reference to the relevant Cynthion. """

        self.device = device



#!/usr/bin/env python3
# pylint: disable=maybe-no-member
#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Generic USB analyzer backend generator for LUNA. """

import time
import errno


import usb
from datetime import datetime
from enum import IntEnum, IntFlag

from amaranth                            import Signal, Elaboratable, Module, DomainRenamer, ResetInserter
from amaranth.build.res                  import ResourceError
from usb_protocol.emitters               import DeviceDescriptorCollection
from usb_protocol.types                  import USBRequestType, USBRequestRecipient

from luna.usb2                           import USBDevice, USBStreamInEndpoint
from luna                                import top_level_cli

from luna.gateware.usb.request.control   import ControlRequestHandler
from luna.gateware.usb.stream            import USBInStreamInterface
from luna.gateware.stream.generator      import StreamSerializer
from luna.gateware.utils.cdc             import synchronize
from luna.gateware.architecture.car      import LunaECP5DomainGenerator
from luna.gateware.architecture.flash_sn import ECP5FlashUIDStringDescriptor
from luna.gateware.interface.ulpi        import UTMITranslator
from luna.gateware.usb.usb2.control      import USBControlEndpoint
from luna.gateware.usb.request.standard  import StandardRequestHandler
from luna.gateware.usb.request.windows   import MicrosoftOS10DescriptorCollection, MicrosoftOS10RequestHandler

from apollo_fpga.gateware.advertiser     import ApolloAdvertiser, ApolloAdvertiserRequestHandler

from usb_protocol.emitters.descriptors.standard import get_string_descriptor
from usb_protocol.types.descriptors.microsoft10 import RegistryTypes

from .analyzer                           import USBAnalyzer
from .fifo                               import Stream16to8, StreamFIFO, AsyncFIFOReadReset, HyperRAMPacketFIFO

import cynthion


USB_SPEED_HIGH       = 0b00
USB_SPEED_FULL       = 0b01
USB_SPEED_LOW        = 0b10

USB_VENDOR_ID        = cynthion.shared.usb.bVendorId.cynthion
USB_PRODUCT_ID       = cynthion.shared.usb.bProductId.cynthion

BULK_ENDPOINT_NUMBER  = 1
BULK_ENDPOINT_ADDRESS = 0x80 | BULK_ENDPOINT_NUMBER
MAX_BULK_PACKET_SIZE  = 512


class USBAnalyzerRegister(Elaboratable):

    def __init__(self, reset=0x00):
        self.current = Signal(8, reset=reset)
        self.next = Signal(8)
        self.write = Signal()

    def elaborate(self, platform):
        m = Module()
        with m.If(self.write):
            m.d.sync += self.current.eq(self.next)
        return m


class USBAnalyzerVendorRequests(IntEnum):
    GET_STATE = 0
    SET_STATE = 1
    GET_SPEEDS = 2
    SET_TEST_CONFIG = 3


class USBAnalyzerSupportedSpeeds(IntFlag):
    USB_SPEED_AUTO = 0b0001
    USB_SPEED_LOW  = 0b0010
    USB_SPEED_FULL = 0b0100
    USB_SPEED_HIGH = 0b1000


class USBAnalyzerVendorRequestHandler(ControlRequestHandler):

    def __init__(self, state, test_config):
        self.state = state
        self.test_config = test_config
        super().__init__()

    def elaborate(self, platform):
        m = Module()
        interface = self.interface

        # Create convenience aliases for our interface components.
        setup               = interface.setup
        handshake_generator = interface.handshakes_out

        # Transmitter for small-constant-response requests
        m.submodules.transmitter = transmitter = \
            StreamSerializer(data_length=1, domain="usb", stream_type=USBInStreamInterface, max_length_width=1)

        # Handle vendor requests to our interface.
        with m.If(
                (setup.type == USBRequestType.VENDOR) &
                (setup.recipient == USBRequestRecipient.INTERFACE) &
                (setup.index == 0)):

            m.d.comb += interface.claim.eq(
                (setup.request == USBAnalyzerVendorRequests.GET_STATE) |
                (setup.request == USBAnalyzerVendorRequests.SET_STATE) |
                (setup.request == USBAnalyzerVendorRequests.GET_SPEEDS)|
                (setup.request == USBAnalyzerVendorRequests.SET_TEST_CONFIG))

            with m.FSM(domain="usb"):

                # IDLE -- not handling any active request
                with m.State('IDLE'):

                    # If we've received a new setup packet, handle it.
                    with m.If(setup.received):

                        # Select which vendor we're going to handle.
                        with m.Switch(setup.request):

                            with m.Case(USBAnalyzerVendorRequests.GET_STATE):
                                m.next = 'GET_STATE'
                            with m.Case(USBAnalyzerVendorRequests.SET_STATE):
                                m.next = 'SET_STATE'
                            with m.Case(USBAnalyzerVendorRequests.GET_SPEEDS):
                                m.next = 'GET_SPEEDS'
                            with m.Case(USBAnalyzerVendorRequests.SET_TEST_CONFIG):
                                m.next = 'SET_TEST_CONFIG'

                # GET_STATE -- Fetch the device's state
                with m.State('GET_STATE'):
                    self.handle_simple_data_request(m, transmitter, self.state.current, length=1)

                # SET_STATE -- The host is trying to set our state
                with m.State('SET_STATE'):
                    self.handle_register_write_request(m, self.state.next, self.state.write)

                # GET_SPEEDS -- Fetch the device's supported USB speeds
                with m.State('GET_SPEEDS'):
                    supported_speeds = \
                        USBAnalyzerSupportedSpeeds.USB_SPEED_LOW | \
                        USBAnalyzerSupportedSpeeds.USB_SPEED_FULL | \
                        USBAnalyzerSupportedSpeeds.USB_SPEED_HIGH
                    self.handle_simple_data_request(m, transmitter, supported_speeds, length=1)

                # SET_TEST_CONFIG -- The host is trying to configure our test device
                with m.State('SET_TEST_CONFIG'):
                    self.handle_register_write_request(m, self.test_config.next, self.test_config.write)

        return m


class USBAnalyzerApplet(Elaboratable):
    """ Gateware that serves as a generic USB analyzer backend.

    WARNING: This is _incomplete_! It's missing:
        - DRAM backing for analysis
    """

    def create_descriptors(self, platform, sharing):
        """ Create the descriptors we want to use for our device. """

        major, minor = platform.version
        descriptors = DeviceDescriptorCollection()

        #
        # We'll add the major components of the descriptors we we want.
        # The collection we build here will be necessary to create a standard endpoint.
        #

        # We'll need a device descriptor...
        with descriptors.DeviceDescriptor() as d:
            d.idVendor           = USB_VENDOR_ID
            d.idProduct          = USB_PRODUCT_ID

            d.iManufacturer      = "Cynthion Project"
            d.iProduct           = "USB Analyzer"
            d.iSerialNumber      = ECP5FlashUIDStringDescriptor
            d.bcdDevice          = major + (minor * 0.01)

            d.bNumConfigurations = 1


        # ... and a description of the USB configuration we'll provide.
        with descriptors.ConfigurationDescriptor() as c:

            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber = 0
                i.bInterfaceClass = 0xFF
                i.bInterfaceSubclass = cynthion.shared.usb.bInterfaceSubClass.analyzer
                i.bInterfaceProtocol = cynthion.shared.usb.bInterfaceProtocol.analyzer

                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = BULK_ENDPOINT_ADDRESS
                    e.wMaxPacketSize   = MAX_BULK_PACKET_SIZE

            # Include Apollo stub interface, if using a shared port.
            if sharing is not None:
                with c.InterfaceDescriptor() as i:
                    i.bInterfaceNumber = 1
                    i.bInterfaceClass = 0xFF
                    i.bInterfaceSubclass = cynthion.shared.usb.bInterfaceSubClass.apollo
                    i.bInterfaceProtocol = ApolloAdvertiserRequestHandler.PROTOCOL_VERSION

        return descriptors


    def elaborate(self, platform):
        m = Module()

        # State register
        m.submodules.state = state = USBAnalyzerRegister()

        # Test config register
        m.submodules.test_config = test_config = USBAnalyzerRegister(reset=0x01)

        # Generate our clock domains.
        clocking = LunaECP5DomainGenerator()
        m.submodules.clocking = clocking

        # Create our UTMI translator.
        ulpi = platform.request("target_phy")
        m.submodules.utmi = utmi = UTMITranslator(ulpi=ulpi)

        # Strap our power controls to be in VBUS passthrough by default,
        # on the target port.
        if platform.version >= (0, 6):
            # On Cynthion r1.4, Target-C to Target-A VBUS passthrough is
            # off by default and must be enabled by the gateware.
            m.d.comb += [
                platform.request("target_c_vbus_en").o  .eq(1),
            ]
            # On Cynthion r0.6 - r1.3 this passthrough is enabled by
            # default, even with the hardware unpowered, but it does no
            # harm to explicitly set it here.
        else:
            # On Cynthion r0.1 - r0.5, there is no `target_c_vbus_en`
            # signal. The following two signals are needed to have
            # the same effect:
            m.d.comb += [
                platform.request("power_a_port").o      .eq(0),
                platform.request("pass_through_vbus").o .eq(1),
            ]

        # Set up our parameters.
        m.d.comb += [

            # Set our mode to non-driving and to the desired speed.
            utmi.op_mode     .eq(0b01),
            utmi.xcvr_select .eq(state.current[1:3]),

            # Disable all of our terminations, as we want to participate in
            # passive observation.
            utmi.dm_pulldown .eq(0),
            utmi.dm_pulldown .eq(0),
            utmi.term_select .eq(0)
        ]

        # Select the appropriate PHY according to platform version.
        if platform.version >= (0, 6):
            phy_name = "control_phy"

            # Also set up a test device on the AUX PHY.
            m.submodules += AnalyzerTestDevice(test_config)
        else:
            phy_name = "host_phy"

        # Check how the port is shared with Apollo.
        sharing = platform.port_sharing(phy_name)

        # Create our USB uplink interface...
        uplink_ulpi = platform.request(phy_name)
        m.submodules.usb = usb = USBDevice(bus=uplink_ulpi)

        # Create descriptors.
        descriptors = self.create_descriptors(platform, sharing)

        # Add Microsoft OS 1.0 descriptors for Windows compatibility.
        descriptors.add_descriptor(get_string_descriptor("MSFT100\xee"), index=0xee)
        msft_descriptors = MicrosoftOS10DescriptorCollection()
        with msft_descriptors.ExtendedCompatIDDescriptor() as c:
            with c.Function() as f:
                f.bFirstInterfaceNumber = 0
                f.compatibleID          = 'WINUSB'
            if sharing is not None:
                with c.Function() as f:
                    f.bFirstInterfaceNumber = 1
                    f.compatibleID          = 'WINUSB'
        with msft_descriptors.ExtendedPropertiesDescriptor() as d:
            with d.Property() as p:
                p.dwPropertyDataType = RegistryTypes.REG_SZ
                p.PropertyName       = "DeviceInterfaceGUID"
                p.PropertyData       = "{88bae032-5a81-49f0-bc3d-a4ff138216d6}"

        # Add our standard control endpoint to the device.
        control_endpoint = usb.add_standard_control_endpoint(descriptors, avoid_blockram=True)

        # Add handler for Microsoft descriptors.
        msft_handler = MicrosoftOS10RequestHandler(msft_descriptors, request_code=0xee)
        control_endpoint.add_request_handler(msft_handler)

        # Add our vendor request handler to the control endpoint.
        vendor_request_handler = USBAnalyzerVendorRequestHandler(state, test_config)
        control_endpoint.add_request_handler(vendor_request_handler)

        # If needed, create an advertiser and add its request handler.
        if sharing == "advertising":
            adv = m.submodules.adv = ApolloAdvertiser()
            control_endpoint.add_request_handler(adv.default_request_handler(1))

        # Add a stream endpoint to our device.
        stream_ep = USBStreamInEndpoint(
            endpoint_number=BULK_ENDPOINT_NUMBER,
            max_packet_size=MAX_BULK_PACKET_SIZE
        )
        usb.add_endpoint(stream_ep)

        # Create a USB analyzer.
        m.submodules.analyzer = analyzer = USBAnalyzer(utmi_interface=utmi)

        # Follow this with a HyperRAM FIFO for additional buffering.
        reset_on_start = ResetInserter(analyzer.discarding)
        m.submodules.psram_fifo = psram_fifo = reset_on_start(
            HyperRAMPacketFIFO(out_fifo_depth=128))

        # Convert the 16-bit stream into an 8-bit one for output.
        m.submodules.s16to8 = s16to8 = reset_on_start(Stream16to8())

        # Add a special stream clock converter for 'sync' to 'usb' crossing.
        m.submodules.clk_conv = clk_conv = StreamFIFO(
            AsyncFIFOReadReset(width=8, depth=4, r_domain="usb", w_domain="sync"))

        m.d.comb += [
            # Connect enable signal to host-controlled state register.
            analyzer.capture_enable     .eq(state.current[0]),

            # Flush endpoint when analyzer is idle with capture disabled.
            stream_ep.flush             .eq(analyzer.idle & ~analyzer.capture_enable),

            # Discard data buffered by endpoint when the analyzer discards its data.
            stream_ep.discard           .eq(analyzer.discarding),

            # USB stream pipeline.
            psram_fifo.input            .stream_eq(analyzer.stream),
            s16to8.input                .stream_eq(psram_fifo.output),
            clk_conv.input              .stream_eq(s16to8.output),
            clk_conv.fifo.ext_rst       .eq(analyzer.discarding),
            stream_ep.stream            .stream_eq(clk_conv.output),

            usb.connect                 .eq(1),

            # LED indicators.
            platform.request("led", 0).o  .eq(analyzer.capturing),
            platform.request("led", 1).o  .eq(stream_ep.stream.valid),
            platform.request("led", 2).o  .eq(analyzer.overrun),

            platform.request("led", 3).o  .eq(utmi.session_valid),
            platform.request("led", 4).o  .eq(utmi.rx_active),
            platform.request("led", 5).o  .eq(utmi.rx_error),
        ]

        # Return our elaborated module.
        return m


class AnalyzerTestDevice(Elaboratable):
    """ Built-in example device that can be used to test the analyzer. """

    SPEEDS = (USB_SPEED_HIGH, USB_SPEED_FULL, USB_SPEED_LOW)

    EP0_MAX_SIZE = {
        USB_SPEED_HIGH: 64,
        USB_SPEED_FULL: 64,
        USB_SPEED_LOW: 8,
    }

    INT_EP_MAX_SIZE = {
        USB_SPEED_HIGH: 512,
        USB_SPEED_FULL: 64,
        USB_SPEED_LOW: 8,
    }

    INT_EP_NUM = {
        USB_SPEED_HIGH: 1,
        USB_SPEED_FULL: 2,
        USB_SPEED_LOW: 3,
    }

    def __init__(self, config):
        self.config = config

    def create_descriptors(self, speed):
        descriptors = DeviceDescriptorCollection()

        with descriptors.DeviceDescriptor() as d:
            d.idVendor           = cynthion.shared.usb.bVendorId.example
            d.idProduct          = cynthion.shared.usb.bProductId.analyzer_test
            d.iManufacturer      = "Cynthion Project"
            d.iProduct           = "USB Analyzer Test Device"
            d.bcdDevice          = 0.01
            d.bNumConfigurations = 1
            d.bMaxPacketSize0    = self.EP0_MAX_SIZE[speed]

        with descriptors.ConfigurationDescriptor() as c:
            with c.InterfaceDescriptor() as i:
                i.bInterfaceNumber = 0
                with i.EndpointDescriptor() as e:
                    e.bEndpointAddress = 0x80 | self.INT_EP_NUM[speed]
                    e.bmAttributes     = 0x03 # Interrupt endpoint
                    e.wMaxPacketSize   = self.INT_EP_MAX_SIZE[speed]
                    e.bInterval        = 0x05 # 5ms interval

        descriptors.add_descriptor(
                get_string_descriptor("MSFT100\xee"), index=0xee)

        return descriptors

    def elaborate(self, platform):
        m = Module()

        # Create a USB device and connect it as required.
        m.submodules.usb = usb = USBDevice(bus=platform.request("aux_phy"))
        current_speed = self.config.current[1:3]
        m.d.comb += [
            usb.connect.eq(self.config.current[0]),
            usb.low_speed_only.eq(current_speed == USB_SPEED_LOW),
            usb.full_speed_only.eq(current_speed == USB_SPEED_FULL),
        ]

        # Create control endpoint.
        control_ep = USBControlEndpoint(utmi=usb.utmi)

        # Add standard request handlers for each speed.
        for speed in self.SPEEDS:
            handler = StandardRequestHandler(
                self.create_descriptors(speed),
                self.EP0_MAX_SIZE[speed],
                avoid_blockram=True,
                blacklist=[lambda setup,speed=speed: current_speed != speed])
            control_ep.add_request_handler(handler)

        # Add Microsoft descriptors for Windows compatibility.
        msft_descriptors = MicrosoftOS10DescriptorCollection()
        with msft_descriptors.ExtendedCompatIDDescriptor() as c:
            with c.Function() as f:
                f.bFirstInterfaceNumber = 0
                f.compatibleID          = 'WINUSB'

        # Add handler for Microsoft descriptors.
        msft_handler = MicrosoftOS10RequestHandler(
                msft_descriptors, request_code=0xee)
        control_ep.add_request_handler(msft_handler)

        # Add control endpoint.
        usb.add_endpoint(control_ep)

        # Add IN endpoints for each speed.
        for speed in self.SPEEDS:
            in_ep = USBStreamInEndpoint(
                endpoint_number=self.INT_EP_NUM[speed],
                max_packet_size=self.INT_EP_MAX_SIZE[speed])
            usb.add_endpoint(in_ep)

            # Output a counter to the endpoint.
            counter = Signal(8)
            m.d.comb += [
                in_ep.stream.valid.eq(1),
                in_ep.stream.payload.eq(counter),
            ]
            with m.If(in_ep.stream.ready):
                m.d.usb += counter.eq(counter + 1)

        return m


if __name__ == "__main__":
    top_level_cli(USBAnalyzerApplet)



import platform

APOLLO_PLATFORMS = {
    (0, 1): 'cynthion.gateware.platform:CynthionPlatformRev0D1',
    (0, 2): 'cynthion.gateware.platform:CynthionPlatformRev0D2',
    (0, 3): 'cynthion.gateware.platform:CynthionPlatformRev0D3',
    (0, 4): 'cynthion.gateware.platform:CynthionPlatformRev0D4',
    (0, 5): 'cynthion.gateware.platform:CynthionPlatformRev0D5',
    (0, 6): 'cynthion.gateware.platform:CynthionPlatformRev0D6',
    (0, 7): 'cynthion.gateware.platform:CynthionPlatformRev0D7',
    (1, 0): 'cynthion.gateware.platform:CynthionPlatformRev1D0',
    (1, 1): 'cynthion.gateware.platform:CynthionPlatformRev1D1',
    (1, 2): 'cynthion.gateware.platform:CynthionPlatformRev1D2',
    (1, 3): 'cynthion.gateware.platform:CynthionPlatformRev1D3',
    (1, 4): 'cynthion.gateware.platform:CynthionPlatformRev1D4',
}



#
# This file is part of Cynthion
#

from ..interface import CynthionInterface


class LED(CynthionInterface):
    """ Simple periheral that allows control of an LED through the Cynthion HAL."""

    def __init__(self, board, led_number):
        """Create a new object representing a Cynthion LED.

        board -- The Cynthion board object that owns the given LED.
        led_number -- The one-indexed LED number. On Cynthion boards, this
                matches the number printed on the silkscreen.
        """

        # Store a reference to the parent board.
        self.board = board

        # Store which of the four(?) LEDs we refer to.
        # TODO: Validate this?
        self.led_number = led_number

    # Function that toggles the relevant LED value. """
    def toggle(self):
        self.board.apis.leds.toggle(self.led_number)

    # Function that turns on the relevant LED value. """
    def on(self):
        self.board.apis.leds.on(self.led_number)

    # Function that turns off the relevant LED value. """
    def off(self):
        self.board.apis.leds.off(self.led_number)



#
# This file is part of Cynthion.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from amaranth               import Cat, Elaboratable, Module, Signal, Record
from luna.gateware.platform import NullPin
from luna_soc.gateware.csr  import Peripheral

from apollo_fpga.gateware   import ApolloAdvertiser

class CynthionInformationPeripheral(Peripheral, Elaboratable):
    """ Peripheral for retrieving Cynthion hardware information at runtime.

    CSR registers
    -------------
    version_major : read
        Cynthion hardware major revision number.
    version_minor : read
        Cynthion hardware minor revision number.
    """

    def __init__(self):
        super().__init__()

        #
        # Registers
        #

        regs         = self.csr_bank()
        self._version_major = regs.csr(8, "r", desc="""
            Contains the Cynthion hardware major revision number.
        """)
        self._version_minor = regs.csr(8, "r", desc="""
            Contains the Cynthion hardware minor revision number.
        """)

        # wishbone connection
        self._bridge = self.bridge(data_width=32, granularity=8, alignment=2)
        self.bus     = self._bridge.bus


    def elaborate(self, platform):
        m = Module()
        m.submodules.bridge = self._bridge

        # Cynthion hardware revision numbers
        major, minor = platform.version
        m.d.comb += [
            self._version_major.r_data.eq(major),
            self._version_minor.r_data.eq(minor),
        ]

        return m



#!/usr/bin/env python3
#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from apollo_fpga.support.selftest import ApolloSelfTestCase, named_test
from .registers import *


ALLOWED_HYPERRAM_IDS = (0x0c81, 0x0c86)


class StandaloneTester(ApolloSelfTestCase):

    def __init__(self, dut):
        self.dut = dut


    def assertPhyRegister(self, phy_register_base: int, register: int, expected_value: int):
        """ Asserts that a PHY register contains a given value.

        Parameters:
            phy_register_base -- The base address of the PHY window in the debug SPI
                                 address range.
            register          -- The PHY register to check.
            value             -- The expected value of the relevant PHY register.
        """

        # Set the address of the ULPI register we're going to read from.
        self.dut.registers.register_write(phy_register_base, register)
        self.dut.registers.register_write(phy_register_base, register)

        # ... and read back its value.
        actual_value = self.dut.registers.register_read(phy_register_base + 1)

        # Finally, validate it.
        if actual_value != expected_value:
            raise AssertionError(f"PHY register {register} was {actual_value}, not expected {expected_value}")


    def assertPhyReadBack(self, phy_register_base: int, value: int):
        """ Writes a value to the PHY scratch register and asserts that the read-back matches.

        Parameters:
            phy_register_base -- The base address of the PHY window in the debug SPI
                                 address range.
            value             -- The value written to the scratch register.
        """

        # Set the address of the ULPI register we're going to read from.
        self.dut.registers.register_write(phy_register_base, 0x16)

        # Write the value to it.
        self.dut.registers.register_write(phy_register_base + 1, value)

        # Set the address again to perform the read.
        self.dut.registers.register_write(phy_register_base, 0x16)

        # ... and read back the value.
        actual_value = self.dut.registers.register_read(phy_register_base + 1)

        # Finally, validate it.
        if actual_value != value:
            raise AssertionError(f"PHY scratch register read-back was {actual_value}, not expected {value}")


    def assertPhyPresence(self, register_base: int):
        """ Assertion that fails iff the given PHY isn't detected. """

        # Check the value of our four ID registers, which should
        # read 2404:0900 (vendor: microchip; product: USB3343).
        self.assertPhyRegister(register_base, 0, 0x24)
        self.assertPhyRegister(register_base, 1, 0x04)
        self.assertPhyRegister(register_base, 2, 0x09)
        self.assertPhyRegister(register_base, 3, 0x00)

        # Write some patterns to the scratch register & read them back
        # to exercise all the DATA# lines.
        self.assertPhyReadBack(register_base, 0x00)
        self.assertPhyReadBack(register_base, 0xff)
        for i in range(8):
            self.assertPhyReadBack(register_base, (1 << i))


    def assertHyperRAMRegister(self, address: int, expected_values: int):
        """ Assertion that fails iff a RAM register doesn't hold the expected value. """

        self.dut.registers.register_write(REGISTER_RAM_REG_ADDR, address)
        actual_value =  self.dut.registers.register_read(REGISTER_RAM_VALUE)

        if actual_value not in expected_values:
            raise AssertionError(f"RAM register {address} was {actual_value}, not one of expected {expected_values}")


    @named_test("Debug module")
    def test_debug_connection(self, dut):
        self.assertRegisterValue(1, 0x54455354)


    @named_test("AUX PHY")
    def test_host_phy(self, dut):
        self.assertPhyPresence(REGISTER_AUX_ADDR)


    @named_test("TARGET PHY")
    def test_target_phy(self, dut):
        self.assertPhyPresence(REGISTER_TARGET_ADDR)


    @named_test("CONTROL PHY")
    def test_sideband_phy(self, dut):
        self.assertPhyPresence(REGISTER_CONTROL_ADDR)


    @named_test("HyperRAM")
    def test_hyperram(self, dut):
        self.assertHyperRAMRegister(0, ALLOWED_HYPERRAM_IDS)



#!/usr/bin/env python3
#
# This file is part of Cynthion.
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

import logging, os, sys, traceback, unittest

import cynthion

class TestSharedValues(unittest.TestCase):
    """Tests for values shared via `cynthion.git/shared/`"""

    def test_usb(self):
        self.assertEqual(cynthion.shared.usb.bVendorId.cynthion, 0x1d50)
        self.assertEqual(cynthion.shared.usb.bProductId.cynthion, 0x615b)

if __name__ == "__main__":
    unittest.main()



#
# This file is part of Cynthion
#

from .board import CynthionBoard

# Ensure that we have access to all Cynthion boards. Normally, we'd avoid
# importing an entire namespace, but in this case, this allows us to ensure
# that all board modules are loaded for autoidentification.
from .boards import *

active_connections = {}


def Cynthion(**board_identifiers):
    """
            Attempts to create a new instance of Cynthion board (sub)class
            most applicable to the given device. For example, if the attached
            board is a Cynthion, this will automatically create a
            Cynthion object.

            Accepts the same arguments as pyusb's usb.find() method, allowing narrowing
            to a more specific Cynthion by e.g. serial number. Like usb.find(), providing
            find_all will return a list of all found devices.

            Throws a DeviceNotFoundError if no device is avaiable and find_all is not set.
    """

    if 'find_all' in board_identifiers and board_identifiers['find_all']:
        del board_identifiers['find_all']
        return CynthionBoard.autodetect_all(**board_identifiers)
    else:
        return CynthionBoard.autodetect(**board_identifiers)


def CynthionSingleton(serial=None):
    """ Returns a Cynthion object, re-using an existing object if we already have a connection to the given Cynthion. """

    # If we already have a Cynthion with the given serial,
    if serial in active_connections:
        device = active_connections[serial]
        if device.comms.still_connected():
            return device

    # Otherwise, try to create a new Cynthion instance.
    Cynthion = Cynthion(serial_number=serial)
    active_connections[serial] = Cynthion


    return Cynthion



#!/usr/bin/env python3
#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

REGISTER_ID = 1
REGISTER_LEDS = 2

REGISTER_TARGET_ADDR = 7
REGISTER_TARGET_VALUE = 8
REGISTER_TARGET_RXCMD = 9

REGISTER_AUX_ADDR = 10
REGISTER_AUX_VALUE = 11
REGISTER_AUX_RXCMD = 12

REGISTER_CONTROL_ADDR = 13
REGISTER_CONTROL_VALUE = 14
REGISTER_CONTROL_RXCMD = 15

REGISTER_RAM_REG_ADDR = 20
REGISTER_RAM_VALUE = 21



#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

import os

from amaranth import Fragment, Module

from amaranth.build import *
try:
    from amaranth.vendor.lattice_ecp5 import LatticeECP5Platform
except:
    from amaranth.vendor import LatticeECP5Platform
from ..vendor.amaranth_boards.resources import *

from luna.gateware.platform.core import LUNAApolloPlatform
from luna.gateware.architecture.car import LunaECP5DomainGenerator

__all__ = ["CynthionPlatform"]

class CynthionPlatform(LUNAApolloPlatform, LatticeECP5Platform):
    """ Board description for Cynthion """

    default_clk = "clk_60MHz"

    # Provide the type that'll be used to create our clock domains.
    clock_domain_generator = LunaECP5DomainGenerator

    #
    # Default clock frequencies for each of our clock domains.
    #
    # Different revisions have different FPGA speed grades, and thus the
    # default frequencies will vary.
    #
    DEFAULT_CLOCK_FREQUENCIES_MHZ = {
        "fast": 240,
        "sync": 120,
        "usb":  60
    }

    #
    # Preferred DRAM bus I/O (de)-skewing constants.
    #
    ram_timings = dict(
        # Set max skew to meet IO setup times
        # TODO: remove this & use the PLL to produce a 90degree clock signal instead.
        clock_skew = 127
    )

    # Provides any platform-specific ULPI registers necessary.
    # This is the spot to put any platform-specific vendor registers that need
    # to be written.
    ulpi_extra_registers = {
        0x39: 0b000110 # USB3343: swap D+ and D- to match the hardware design
    }

    def toolchain_prepare(self, fragment, name, **kwargs):
        overrides = {
            'ecppack_opts': '--compress --freq 38.8'
        }

        return super().toolchain_prepare(fragment, name, **overrides, **kwargs)

    def toolchain_program(self, products, name):
        """ Programs the relevant LUNA board via its sideband connection. """

        from apollo_fpga import ApolloDebugger
        from apollo_fpga.ecp5 import ECP5_JTAGProgrammer

        # Create our connection to the debug module.
        debugger = ApolloDebugger()

        # Grab our generated bitstream, and upload it to the FPGA.
        bitstream =  products.get("{}.bit".format(name))
        with debugger.jtag as jtag:
            programmer = ECP5_JTAGProgrammer(jtag)
            programmer.configure(bitstream)

        # Let the LUNA gateware take over in devices with shared USB port
        try:
            debugger.allow_fpga_takeover_usb()
        except AttributeError:
            pass


    def toolchain_flash(self, products, name="top"):
        """ Programs the LUNA board's flash via its sideband connection. """

        from apollo_fpga import ApolloDebugger
        from apollo_fpga.ecp5 import ECP5_JTAGProgrammer

        # Create our connection to the debug module.
        debugger = ApolloDebugger()
        self._ensure_unconfigured(debugger)

        # Grab our generated bitstream, and upload it to the .
        bitstream =  products.get("{}.bit".format(name))
        with debugger.jtag as jtag:
            programmer = ECP5_JTAGProgrammer(jtag)
            programmer.flash(bitstream)

        debugger.soft_reset()

        # Let the LUNA gateware take over in devices with shared USB port
        try:
            debugger.allow_fpga_takeover_usb()
        except AttributeError:
            pass

    def pseudo_power_supply_fragment(self):
        """ Fragment to assign fixed values to the pseudo power supply pins """
        m = Module()
        if ("pseudo_vccio", 0) in self.resources:
            m.d.comb += self.request("pseudo_vccio").o.eq(-1)
        if ("pseudo_gnd", 0) in self.resources:
            m.d.comb += self.request("pseudo_gnd").o.eq(0)
        return m

    # This list can be safely overriden in child classes.
    # Fragments defined in this class are always added.
    append_fragments = [ pseudo_power_supply_fragment ]

    def prepare(self, elaboratable, name="top", **kwargs):
        fragment = Fragment.get(elaboratable, self)

        # Merge base CynthionPlatform fragments with board-specific ones
        append_fragments = set(CynthionPlatform.append_fragments) | set(self.append_fragments)

        for subfragment in append_fragments:
            fragment.add_subfragment(Fragment.get(subfragment(self), self))

        return super().prepare(fragment, name, **kwargs)



#!/usr/bin/env python
#
# This file is part of Cynthion.
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Cynthion 'flash' command. """

import logging

from .util import find_cynthion_asset, find_cynthion_bitstream
from .util import flash_bitstream, flash_mcu_firmware, flash_soc_firmware


def cynthion_flash(device, args):
    if args.bitstream is not None:
        flash_bitstream(device, args.bitstream)
    elif args.mcu_firmware is not None:
        flash_mcu_firmware(device, args.mcu_firmware)
    elif args.soc_firmware is not None:
        flash_soc_firmware(device, args.soc_firmware)
    elif args.target == "analyzer":
        flash_bitstream(device, find_cynthion_bitstream(device, "analyzer.bit"))
    elif args.target == "facedancer":
        flash_soc_firmware(device, find_cynthion_asset("moondancer.bin"))
        flash_bitstream(device, find_cynthion_bitstream(device, "facedancer.bit"))
    elif args.target == "selftest":
        flash_bitstream(device, find_cynthion_bitstream(device, "selftest.bit"))
    else:
        logging.error(f"unknown target: {args.target}")
        return



#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

import os

from amaranth.build import *
from ..vendor.amaranth_boards.resources import *

from .core import CynthionPlatform

__all__ = ["CynthionPlatformRev0D2"]

#
# Note that r0.2+ have D+/D- swapped to avoid having to cross D+/D- in routing.
#
# This is supported by a PHY feature that allows you to swap pins 13 + 14.
#

class CynthionPlatformRev0D2(CynthionPlatform):
    """ Board description for the pre-release r0.2 revision of Cynthion. """

    name        = "Cynthion r0.2"
    version     = (0, 2)
    device      = "LFE5U-12F"
    package     = "BG256"
    speed       = os.getenv("ECP5_SPEED_GRADE", "8")

    # By default, assume we'll be connecting via our target PHY.
    default_usb_connection = "host_phy"

    #
    # Preferred DRAM bus I/O (de)-skewing constants.
    #
    ram_timings = dict(
        clock_skew = 64
    )

    #
    # I/O resources.
    #
    resources   = [

        # Primary, discrete 60MHz oscillator.
        Resource("clk_60MHz", 0, Pins("A7", dir="i"),
            Clock(60e6), Attrs(IO_TYPE="LVCMOS33")),

        # Connection to our SPI flash; can be used to work with the flash
        # from e.g. a bootloader.
        Resource("spi_flash", 0,

            # SCK is on pin 9; but doesn't have a traditional I/O buffer.
            # Instead, we'll need to drive a clock into a USRMCLK instance.
            # See interfaces/flash.py for more information.
            Subsignal("sdi",  Pins("T8",  dir="o")),
            Subsignal("sdo",  Pins("T7",  dir="i")),

            # In r0.2, the chip select line can either be driven by the FPGA
            # or by the Debug Controller. Accordingly, we'll mark the line as
            # bidirectional, and let the user decide.
            Subsignal("cs",   PinsN("N8", dir="io")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # Connection to our SPI flash but using quad mode (QSPI)
        Resource("qspi_flash", 0,
            # SCK is on pin 9; but doesn't have a traditional I/O buffer.
            # Instead, we'll need to drive a clock into a USRMCLK instance.
            # See interfaces/flash.py for more information.
            Subsignal("dq",  Pins("T8 T7 M7 N7",  dir="io")),
            Subsignal("cs",  PinsN("N8", dir="o")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # UART connected to the debug controller; can be routed to a host via CDC-ACM.
        UARTResource(0, rx="R14", tx="T14", attrs=Attrs(IO_TYPE="LVCMOS33")),

        # SPI bus connected to the debug controller, for simple register exchanges.
        # Note that the Debug Controller is the controller on this bus.
        Resource("debug_spi", 0,
            Subsignal("sck",  Pins( "R13", dir="i")),
            Subsignal("sdi",  Pins( "P13", dir="i")),
            Subsignal("sdo",  Pins( "P11", dir="o")),
            Subsignal("cs",   PinsN("T13", dir="i")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # FPGA-connected LEDs.
        *LEDResources(pins="L16 L15 M16 M15 N16 P15", attrs=Attrs(IO_TYPE="LVCMOS33"), invert=True),

        # USB PHYs
        ULPIResource("target_phy", 0,
            data="D16 E15 E16 F15 F16 G15 J16 K16", clk="B15", clk_dir='o',
            dir="C15", nxt="C16", stp="B16", rst="G16", rst_invert=True,
            attrs=Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
        ULPIResource("sideband_phy", 0,
            data="R2 R1 P2 P1 N1 M2 M1 L2", clk="R4", clk_dir='o',
            dir="T3", nxt="T2", stp="T4", rst="R3", rst_invert=True,
            attrs=Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
        ULPIResource("host_phy", 0,
            data="G2 G1 F2 F1 E1 D1 C1 B1", clk="K2", clk_dir='o',
            dir="J1", nxt="H2", stp="J2", rst="K1", rst_invert=True,
            attrs=Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),

        # Target port power switching.
        Resource("power_a_port",       0, Pins("C14", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("pass_through_vbus",  0, Pins("D14", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("target_vbus_fault",  0, Pins("K15", dir="i"), Attrs(IO_TYPE="LVCMOS33")),

        # HyperRAM (1V8 domain).
        Resource("ram", 0,
            # Note: our clock uses the pseudo-differential I/O present on the top tiles.
            # This requires a recent version of trellis+nextpnr. If your build complains
            # that LVCMOS18D is an invalid I/O type, you'll need to upgrade.
            Subsignal("clk",   DiffPairs("B14", "A15", dir="o"), Attrs(IO_TYPE="LVCMOS18D")),
            Subsignal("dq",    Pins("A11 B10 B12 A12 B11 A10 B9 A9", dir="io")),
            Subsignal("rwds",  Pins( "A13", dir="io")),
            Subsignal("cs",    PinsN("A14", dir="o")),
            Subsignal("reset", PinsN("B13", dir="o")),
            Attrs(IO_TYPE="LVCMOS18", SLEWRATE="FAST")
        ),

        # User I/O connections.
        Resource("user_io", 0, Pins("1", conn=("user_io", 0), dir="io"), Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
        Resource("user_io", 1, Pins("3", conn=("user_io", 0), dir="io"), Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
        Resource("user_io", 2, Pins("4", conn=("user_io", 0), dir="io"), Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
        Resource("user_io", 3, Pins("6", conn=("user_io", 0), dir="io"), Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
    ]

    connectors = [
        Connector("user_io", 0, """
            A5  -  A2
            A4  -  A3
        """)
    ]



#!/usr/bin/env python
#
# This file is part of Cynthion.
#
# Copyright (c) 2023-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Root command that delegates to all Cynthion subcommands. """

import argparse
import logging
import sys

from sys                       import platform

from apollo_fpga               import ApolloDebugger

from .util                     import HelpFormatter

from . import cynthion_info, cynthion_flash, cynthion_build, cynthion_run, cynthion_setup, cynthion_update


def main():
    # Set up python's logging to act as a simple print, for now.
    logging.basicConfig(level=logging.INFO, format="%(message)-s")

    # Set up an argument parser.
    parser = argparse.ArgumentParser(description="Cynthion command line interface",
                                     formatter_class=HelpFormatter)
    command_parsers = parser.add_subparsers(dest="command", metavar="command")

    # cynthion run
    run_parser = command_parsers.add_parser("run", help="run a bitstream on the FPGA", formatter_class=HelpFormatter)
    run_parser.set_defaults(func=cynthion_run)
    run_parser.add_argument("--bitstream", help="(advanced) run the bitstream at <filename>", metavar="<filename>", required=False)
    run_subparsers  = run_parser.add_subparsers(dest="target", metavar="BITSTREAM")
    run_analyzer_parser = run_subparsers.add_parser("analyzer", help="run the USB Analyzer bitstream", formatter_class=HelpFormatter)
    run_facedancer_parser = run_subparsers.add_parser("facedancer", help="run the Facedancer bitstream", formatter_class=HelpFormatter)
    run_selftest_parser = run_subparsers.add_parser("selftest", help="run the hardware self-test bitstream", formatter_class=HelpFormatter)

    # cynthion flash
    flash_parser = command_parsers.add_parser("flash", help="overwrite the FPGA's configuration flash with the target bitstream", formatter_class=HelpFormatter)
    flash_parser.set_defaults(func=cynthion_flash)
    flash_group = flash_parser.add_mutually_exclusive_group()
    flash_group.add_argument("--bitstream", help="(advanced) flash the bitstream at <filename>", metavar="<filename>", required=False)
    flash_group.add_argument("--soc-firmware", help="(advanced) flash the soc firmware at <filename>", metavar="<filename>", required=False)
    flash_group.add_argument("--mcu-firmware", help="(advanced) flash the mcu firmware at <filename>", metavar="<filename>", required=False)
    flash_subparsers  = flash_parser.add_subparsers(dest="target", metavar="BITSTREAM")
    flash_analyzer_parser = flash_subparsers.add_parser("analyzer", help="flash the USB Analyzer bitstream", formatter_class=HelpFormatter)
    flash_facedancer_parser = flash_subparsers.add_parser("facedancer", help="flash the Facedancer bitstream", formatter_class=HelpFormatter)

    # cynthion build
    build_parser = command_parsers.add_parser("build")
    build_parser.set_defaults(func=cynthion_build)
    build_parser.add_argument("--luna-platform", action='store_true', help="display the LUNA_PLATFORM variable for the connected device")
    build_subparsers  = build_parser.add_subparsers(dest="target", metavar="TARGET")
    build_analyzer_parser = build_subparsers.add_parser("analyzer")
    build_facedancer_parser = build_subparsers.add_parser("facedancer")
    build_selftest_parser = build_subparsers.add_parser("selftest")

    # cynthion update
    update_parser = command_parsers.add_parser("update", help="update MCU firmware and FPGA configuration flash to the latest installed versions", formatter_class=HelpFormatter)
    update_parser.set_defaults(func=cynthion_update)
    update_parser.add_argument("--mcu-firmware", action='store_true',   help="only update the MCU firmware")
    update_parser.add_argument("--bitstream", action='store_true', help="only update the FPGA bitstream")

    # cynthion info
    info_parser = command_parsers.add_parser("info", help="print device information", formatter_class=HelpFormatter)
    info_parser.add_argument("--force-offline", action='store_true', help="force the FPGA to release the CONTROL port")
    info_parser.set_defaults(func=cynthion_info)

    # cynthion setup (Linux-only for now)
    if platform == "linux" or platform == "linux2":
        setup_parser = command_parsers.add_parser("setup", help="install Cynthion support files required for operation", formatter_class=HelpFormatter)
        setup_parser.set_defaults(func=cynthion_setup)
        setup_group = setup_parser.add_mutually_exclusive_group()
        setup_group.add_argument("--check", action='store_true', help="check Cynthion support files")
        setup_group.add_argument("--uninstall", action='store_true', help="remove Cynthion support files")
        setup_parser.add_argument("--udev", action='store_true', help="Linux udev access rules")

    # Parse arguments.
    args = parser.parse_args()
    if not args.command:
        parser.print_help()
        return

    # Force the FPGA offline by default in most commands to force Apollo mode if needed.
    force_offline = args.force_offline if "force_offline" in args else True

    # Execute the relevant command.
    if args.func in (cynthion_info, cynthion_setup):
        args.func(args)
    else:
        device = ApolloDebugger(force_offline=force_offline)
        args.func(device, args)



#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

import os

from amaranth.build import *
from ..vendor.amaranth_boards.resources import *

from .core import CynthionPlatform

__all__ = ["CynthionPlatformRev0D6"]

class CynthionPlatformRev0D6(CynthionPlatform):
    """ Board description for Cynthion r0.6 """

    name        = "Cynthion r0.6"
    version     = (0, 6)
    device      = "LFE5U-12F"
    package     = "BG256"
    speed       = os.getenv("ECP5_SPEED_GRADE", "8")

    # By default, assume we'll be connecting via our control PHY.
    default_usb_connection = "aux_phy"

    #
    # Preferred DRAM bus I/O (de)-skewing constants.
    #
    ram_timings = dict(
        # Set max skew to meet IO setup times
        # TODO: remove this & use the PLL to produce a 90degree clock signal instead.
        clock_skew = 127
    )

    #
    # I/O resources.
    #
    resources   = [

        # Primary, discrete 60MHz oscillator.
        Resource("clk_60MHz", 0, Pins("A8", dir="i"),
            Clock(60e6), Attrs(IO_TYPE="LVCMOS33")),

        # Connection to our SPI flash; can be used to work with the flash
        # from e.g. a bootloader.
        Resource("spi_flash", 0,

            # SCK is on pin 9; but doesn't have a traditional I/O buffer.
            # Instead, we'll need to drive a clock into a USRMCLK instance.
            # See interfaces/flash.py for more information.
            Subsignal("sdi",  Pins("T8",  dir="o")),
            Subsignal("sdo",  Pins("T7",  dir="i")),
            Subsignal("cs",   PinsN("N8", dir="o")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # Connection to our SPI flash but using quad mode (QSPI)
        Resource("qspi_flash", 0,
            # SCK is on pin 9; but doesn't have a traditional I/O buffer.
            # Instead, we'll need to drive a clock into a USRMCLK instance.
            # See interfaces/flash.py for more information.
            Subsignal("dq",  Pins("T8 T7 M7 N7",  dir="io")),
            Subsignal("cs",  PinsN("N8", dir="o")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # Note: UART pins R14 and T14 are connected to JTAG pins R11 (TDI)
        # and T11 (TMS) respectively, so the microcontroller can use either
        # function but not both simultaneously.

        # UART connected to the debug controller; can be routed to a host via CDC-ACM.
        Resource("uart", 0,
            Subsignal("rx",  Pins("R14",  dir="i")),
            Subsignal("tx",  Pins("T14",  dir="oe"), Attrs(PULLMODE="UP")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # interrupt output to send signal to microcontroller
        Resource("int", 0, Pins("R8", dir="o"), Attrs(IO_TYPE="LVCMOS33")),

        # USER button
        Resource("button_user", 0, PinsN("M14", dir="i"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP")),

        # output signal connected to PROGRAMN to trigger FPGA reconfiguration
        Resource("self_program", 0, PinsN("T13", dir="o"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP")),

        # FPGA LEDs
        *LEDResources(pins="E13 C13 B14 A15 D12 C11", attrs=Attrs(IO_TYPE="LVCMOS33"), invert=True),

        # USB PHYs
        ULPIResource("control_phy", 0,
            data="N16 N14 P16 P15 R16 R15 T15 P14", clk="L14", clk_dir='o',
            dir="M16", nxt="M15", stp="L15", rst="L16", rst_invert=True,
            attrs=Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
        ULPIResource("aux_phy", 0,
            data="J16 K15 K16 J13 J14 H13 H14 K14", clk="F16", clk_dir='o',
            dir="H15", nxt="J15", stp="G16", rst="G15", rst_invert=True,
            attrs=Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
        ULPIResource("target_phy", 0,
            data="R2 R1 P2 P1 N3 N1 M2 M1", clk="T4", clk_dir='o',
            dir="R3", nxt="T2", stp="T3", rst="R4", rst_invert=True,
            attrs=Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),

        # direct connection to TARGET USB D+/D-
        Resource("target_usb_diff", 0, DiffPairs("N4", "P3", dir="i"), Attrs(IO_TYPE="LVDS", PULLMODE="NONE")),
        Resource("target_usb_dp", 0, Pins("N4", dir="i"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="NONE")),
        Resource("target_usb_dm", 0, Pins("P3", dir="i"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="NONE")),

        # USB Type-C controllers and pins
        Resource("target_type_c", 0,
            Subsignal("scl",   Pins( "A4", dir="o" ), Attrs(PULLMODE="NONE")),
            Subsignal("sda",   Pins( "C4", dir="io"), Attrs(PULLMODE="NONE")),
            Subsignal("int",   PinsN("A3", dir="i" ), Attrs(PULLMODE="UP")),
            Subsignal("fault", PinsN("D4", dir="i" ), Attrs(PULLMODE="UP")),
            Subsignal("sbu1",  Pins( "A2", dir="io")),
            Subsignal("sbu2",  Pins( "E4", dir="io")),
            Attrs(IO_TYPE="LVCMOS33")
        ),
        Resource("aux_type_c", 0,
            Subsignal("scl",   Pins( "D16", dir="o" ), Attrs(PULLMODE="NONE")),
            Subsignal("sda",   Pins( "E15", dir="io"), Attrs(PULLMODE="NONE")),
            Subsignal("int",   PinsN("H12", dir="i" ), Attrs(PULLMODE="UP")),
            Subsignal("fault", PinsN("G14", dir="i" ), Attrs(PULLMODE="UP")),
            Subsignal("sbu1",  Pins( "E16", dir="io")),
            Subsignal("sbu2",  Pins( "F15", dir="io")),
            Attrs(IO_TYPE="LVCMOS33")
        ),

        # power input shutoff
        Resource("control_vbus_in_en", 0, PinsN("K13", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("aux_vbus_in_en",     0, PinsN("L13", dir="o"), Attrs(IO_TYPE="LVCMOS33")),

        # VBUS passthrough
        #
        # VBUS on each of the Type-C ports can be connected to TARGET A through
        # a bidirectional switch. If any of these switches is enabled, TARGET A
        # is considered an output. An additional switch can be enabled to pass
        # VBUS through to another port in addition to TARGET A.
        #
        # The TARGET C switch is enabled by default, even when Cynthion is
        # powered off, enabling VBUS passthrough from TARGET C to TARGET A.

        Resource("target_c_vbus_en",   0, PinsN("K5", dir="o"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP")),
        Resource("control_vbus_en",    0, Pins("L1", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("aux_vbus_en",        0, Pins("L2", dir="o"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("target_a_discharge", 0, Pins("K4", dir="o"), Attrs(IO_TYPE="LVCMOS33")),

        # voltage and current monitor
        Resource("power_monitor", 0,
            Subsignal("scl",   Pins( "C6", dir="o" ), Attrs(PULLMODE="NONE")),
            Subsignal("sda",   Pins( "D6", dir="io"), Attrs(PULLMODE="NONE")),
            Subsignal("pwrdn", PinsN("C7", dir="o" )),
            Subsignal("slow",  Pins( "D5", dir="io")),
            Subsignal("gpio",  Pins( "D7", dir="io")),
            Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP")
        ),

        # HyperRAM
        Resource("ram", 0,
            Subsignal("clk",   DiffPairs("C3", "D3", dir="o"), Attrs(IO_TYPE="LVCMOS33D")),
            Subsignal("dq",    Pins("F2 B1 C2 E1 E3 E2 F3 G4", dir="io")),
            Subsignal("rwds",  Pins( "D1", dir="io")),
            Subsignal("cs",    PinsN("B2", dir="o")),
            Subsignal("reset", PinsN("C1", dir="o")),
            Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")
        ),

        # User I/O connections.
        Resource("user_pmod", 0, Pins("1 2 3 4 7 8 9 10", conn=("pmod", 0), dir="io"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("user_pmod", 1, Pins("1 2 3 4 7 8 9 10", conn=("pmod", 1), dir="io"), Attrs(IO_TYPE="LVCMOS33")),
        Resource("user_mezzanine", 0,
                Pins("3 4 5 6 7 8 9 10 11 12 13 18 19 20 21 22 23 24 25 26 27 28", conn=("mezzanine", 0), dir="io"),
                Attrs(IO_TYPE="LVCMOS33", SLEWRATE="FAST")),
    ]

    connectors = [
        Connector("pmod", 0, "C9 C8 D11 C12 - - B8 D8 D9 C10 - -"), # PMOD A
        Connector("pmod", 1, "B4 B5 B6 B7 - - C5 A5 A6 A7 - -"), # PMOD B
        Connector("mezzanine", 0,
            "- - B9 A9 B10 A10 B11 D14 C14 F14 E14 G13 G12 - - - - C16 C15 B16 B15 A14 B13 A13 D13 A12 B12 A11 - -"),
    ]

    apollo_port_sharing = {'control_phy': 'advertising'}



#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from .cynthion_r0_1 import CynthionPlatformRev0D1
from .cynthion_r0_2 import CynthionPlatformRev0D2
from .cynthion_r0_3 import CynthionPlatformRev0D3
from .cynthion_r0_4 import CynthionPlatformRev0D4
from .cynthion_r0_5 import CynthionPlatformRev0D5
from .cynthion_r0_6 import CynthionPlatformRev0D6
from .cynthion_r0_7 import CynthionPlatformRev0D7
from .cynthion_r1_0 import CynthionPlatformRev1D0
from .cynthion_r1_1 import CynthionPlatformRev1D1
from .cynthion_r1_2 import CynthionPlatformRev1D2
from .cynthion_r1_3 import CynthionPlatformRev1D3
from .cynthion_r1_4 import CynthionPlatformRev1D4



#!/usr/bin/env python
#
# This file is part of Cynthion.
#
# Copyright (c) 2023-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Cynthion 'info' command. """

import logging, sys

from apollo_fpga import ApolloDebugger

from .. import __version__, shared


def cynthion_info(args):
    logging.info(f"Cynthion version: {__version__}")
    if ApolloDebugger.print_info(force_offline=args.force_offline, out=logging.info,
                ids=[(shared.usb.bVendorId.apollo, shared.usb.bProductId.apollo)],
                stub_ids=[(shared.usb.bVendorId.cynthion, shared.usb.bProductId.cynthion)]):
        if not args.force_offline:
            logging.info("For additional device information use the --force-offline option.")
    else:
        logging.info("Could not find a Cynthion device.")
        sys.exit(1)



#!/usr/bin/env python
#
# This file is part of Cynthion.
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Cynthion 'setup' command. """

import logging, os, shutil, subprocess, sys

from .util import find_cynthion_asset

UDEV_PATH     = "/etc/udev/rules.d"
UDEV_CYNTHION = "54-cynthion.rules"


def cynthion_setup(args):
    if args.check:
        _check_udev(args)
    elif args.uninstall:
        _uninstall_udev(args)
    else:
        _install_udev(args)


def _check_udev(args):
    logging.info("Checking: Linux udev rules")

    sys_path = os.path.join(UDEV_PATH, UDEV_CYNTHION)

    if not os.path.isfile(sys_path):
        logging.error(f"❌\t{UDEV_CYNTHION} not installed")
        logging.info("\nPlease run 'cynthion setup' to install.")
        sys.exit(1)

    logging.info(f"✅\t{UDEV_CYNTHION} is present")

    sys_rules     = open(sys_path, "r").readlines()
    factory_rules = open(find_cynthion_asset(UDEV_CYNTHION), "r").readlines()

    if sys_rules != factory_rules:
        logging.error(f"❌\t{UDEV_CYNTHION} differs from factory rules")
        logging.info("\nPlease run 'cynthion setup' to re-install.")
        sys.exit(1)

    logging.info(f"✅\t{UDEV_CYNTHION} matches factory rules")
    logging.info("\nAll checks completed successfully.")


def _install_udev(args):
    logging.info("Installing: Linux udev rules")

    src = find_cynthion_asset(UDEV_CYNTHION)
    dst = os.path.join(UDEV_PATH, UDEV_CYNTHION)

    # copy udev rules file
    _run_shell_command(f"cp {src} {dst}", root=True)

    # reload udev rules
    _run_shell_command("udevadm control --reload", root=True)

    # apply udev rules to any devices that are already plugged in
    _run_shell_command("udevadm trigger", root=True)

    logging.info("\nInstallation completed successfully.")


def _uninstall_udev(args):
    logging.info("Uninstalling: Linux udev rules")

    rules = os.path.join(UDEV_PATH, UDEV_CYNTHION)

    if os.path.isfile(rules):
        # remove udev rules file
        _run_shell_command(f"rm {rules}", root=True)

        # reload udev rules
        _run_shell_command("udevadm control --reload", root=True)

        # apply udev rules to any devices that are already plugged in
        _run_shell_command("udevadm trigger", root=True)
    else:
        logging.info(f"✅\t{rules} not present, skipping.")

    logging.info("\nUninstallation completed successfully.")


def _run_shell_command(cmd, root=False):

    if root and os.getuid() != 0:
        SUDO_PATH = shutil.which('sudo')
        if SUDO_PATH is None:
            raise OSError('Cannot find sudo executable.')
        cmd = f"sudo {cmd}"
        
    proc = subprocess.Popen(args=cmd.split())
    proc.wait()

    if proc.returncode != 0:
        logging.error(f"❌\t{cmd} failed with exit code: {proc.returncode}")
        sys.exit(proc.returncode)

    logging.info(f"✅\t{cmd}")



#!/usr/bin/env python
#
# This file is part of Cynthion.
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Cynthion 'build' command. """

import logging, os, subprocess, sys, tempfile

from . import util


def cynthion_build(device, args):
    if args.luna_platform:
        platform_name = util._get_appropriate_platform_name(device)
        sys.stdout.write(f"cynthion.gateware.platform:{platform_name}\n")
        sys.exit(0)

    elif args.target == "analyzer":
        _build_analyzer(device, args)
    elif args.target == "facedancer":
        _build_facedancer(device, args)
    elif args.target == "selftest":
        _build_selftest(device, args)
    else:
        logging.error(f"unknown target: {args.target}")


def _build_analyzer(device, args):
    from luna import top_level_cli
    from cynthion.gateware.analyzer.top import USBAnalyzerApplet
    sys.argv = [sys.argv[0]]
    top_level_cli(USBAnalyzerApplet)


def _build_facedancer(device, args):
    from luna_soc import top_level_cli
    from cynthion.gateware.facedancer.top import MoondancerSoc

    with tempfile.TemporaryDirectory() as tmp:
        filename = os.path.join(tmp, "moondancer.bin")
        proc = subprocess.Popen(
            args=["cargo", "objcopy", "--release", "--", "-Obinary", filename],
            cwd="../../firmware/moondancer/"
        )
        proc.wait()
        flash_soc_firmware(device, filename)

    sys.argv = [sys.argv[0]]
    top_level_cli(MoondancerSoc())


def _build_selftest(device, args):
    from cynthion.selftest.gateware import SelftestDevice
    from cynthion.selftest.host import StandaloneTester

    from luna import top_level_cli
    sys.argv = [sys.argv[0]]
    device = top_level_cli(SelftestDevice)

    # run test
    tester = StandaloneTester(device)
    tester.run_tests()
    print()



#!/usr/bin/env python3
#
# This file is part of Cynthion
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Utility to run a hardware self-test on a connected Cynthion. """

from cynthion.selftest.gateware import SelftestDevice
from cynthion.selftest.host import StandaloneTester
from luna import top_level_cli


def main():
    device = top_level_cli(SelftestDevice)

if __name__ == '__main__':
    main()



from os.path import dirname, basename, isfile
import glob

# Autodetect all boards in the boards directory, so they can be automatically
# imported by 'from boards import *'. This allows easy development of new
# boards-- just create your boards/ file in boards, and be sure to subclass
# CynthionBoard.
modules = glob.glob(dirname(__file__)+"/*.py")
__all__ = [ basename(f)[:-3] for f in modules if isfile(f)]



#
# This file is part of Cynthion
#

class DeviceNotFoundError(IOError):
    """ Error indicating no Cynthion device was found. """
    pass


class DeviceBusyError(IOError):
    """ Error indicating the Cynthion is too busy to service the given request. """
    pass


class DeviceMemoryError(MemoryError):
    """ Error indicating that the Cynthion has run out of memory. """
    pass


class NotFoundError(IOError):
    """ Error indicating that a resource was not found. """
    pass

class CynthionError(RuntimeError):
    """ Runtime error used when no better description is available. """
    pass


class ExternalDeviceError(IOError):
    """
    Error used when a external device (e.g. not on the Cynthion)
    experiences an issue. This typically means that the error is not with
    the Cynthion hardware or software, but may be with e.g. connections.
    """


CYNTHION_ERRORS = {
    -2: ValueError,
    -5: NotFoundError,
    -6: DeviceBusyError,
    -7: MemoryError,
}


def from_Cynthion_error(error_number):
    """
    Returns the error class appropriate for the given Cynthion error.
    """
    error_class = CYNTHION_ERRORS.get(error_number, CynthionError)
    message = "Error {}".format(error_number)
    return error_class(message)



#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Low-level USB analyzer gateware. """

import unittest

from amaranth          import Signal, Module, Elaboratable, Memory, Record, Mux, Cat, C
from enum              import IntEnum

from luna.gateware.stream import StreamInterface
from luna.gateware.test   import LunaGatewareTestCase, usb_domain_test_case

from .fifo import Stream16to8, StreamFIFO, AsyncFIFOReadReset


class USBAnalyzer(Elaboratable):
    """ Core USB analyzer; backed by a small ringbuffer in FPGA block RAM.

    If you're looking to instantiate a full analyzer, you'll probably want to grab
    one of the DRAM-based ringbuffer variants (which are currently forthcoming).

    If you're looking to use this with a ULPI PHY, rather than the FPGA-convenient UTMI interface,
    grab the UTMITranslator from `luna.gateware.interface.ulpi`.

    Attributes
    ----------
    stream: StreamInterface(), output stream
        Stream that carries USB analyzer data.

    idle: Signal(), output
        Asserted iff the analyzer is not currently receiving data.
    stopped: Signal(), output
        Asserted iff the analyzer is stopped and not capturing packets.
    overrun: Signal(), output
        Asserted iff the analyzer has received more data than it can store in its internal buffer.
        Occurs if :attr:``stream`` is not being read quickly enough.
    capturing: Signal(), output
        Asserted iff the analyzer is currently capturing a packet.
    discarding: Signal(), output
        Asserted iff the analyzer is discarding the contents of its internal buffer.


    Parameters
    ----------
    utmi_interface: UTMIInterface()
        The UTMI interface that carries the data to be analyzed.
    mem_depth: int, default=8192
        The depth of the analyzer's local ringbuffer, in bytes.
        Must be a power of 2.
    """

    # Header is 16-bit length and 16-bit timestamp.
    HEADER_SIZE_BYTES = 4
    HEADER_SIZE_WORDS = HEADER_SIZE_BYTES // 2

    # Event is 0xFF marker, 8-bit event code and 16-bit timestamp.
    EVENT_SIZE_BYTES = 4
    EVENT_SIZE_WORDS = HEADER_SIZE_BYTES // 2

    # Support a maximum payload size of 1024B, plus a 1-byte PID and a 2-byte CRC16.
    MAX_PACKET_SIZE_BYTES = 1024 + 1 + 2

    def __init__(self, *, utmi_interface, mem_depth=4096):
        """
        Parameters:
            utmi_interface -- A record or elaboratable that presents a UTMI interface.
        """

        self.utmi = utmi_interface

        assert (mem_depth % 2) == 0, "mem_depth must be a power of 2"

        # Internal storage memory.
        self.mem = Memory(width=16, depth=mem_depth, name="analysis_ringbuffer")
        self.mem_size_words = mem_depth
        self.mem_size_bytes = 2 * mem_depth

        #
        # I/O port
        #
        self.stream         = StreamInterface(payload_width=16)

        self.capture_enable = Signal()
        self.idle           = Signal()
        self.stopped        = Signal()
        self.overrun        = Signal()
        self.capturing      = Signal()
        self.discarding     = Signal()


    def elaborate(self, platform):
        m = Module()

        # Memory read and write ports.
        m.submodules.read  = mem_read_port  = self.mem.read_port(domain="sync", transparent=False)
        m.submodules.write = mem_write_port = self.mem.write_port(domain="sync", granularity=8)

        # FIFO write addresses point to bytes.
        write_byte_addr  = Signal(range(self.mem_size_bytes))

        # Memory addresses point to words
        header_word_addr = Signal.like(mem_write_port.addr)
        write_word_addr  = Signal.like(mem_write_port.addr)
        read_word_addr   = Signal.like(mem_read_port.addr)
        fifo_word_count  = Signal.like(mem_read_port.addr)
        write_odd        = Signal()
        m.d.comb += Cat(write_odd, write_word_addr).eq(write_byte_addr)
        next_word_addr   = (write_byte_addr + write_odd)[1:]

        # Current receive status.
        packet_size     = Signal(16)
        packet_time     = Signal(16)
        event_code      = Signal(8)

        # Triggers for memory write operations.
        write_packet    = Signal()
        write_header    = Signal()
        write_event     = Signal()

        # Use the FIFO as our stream source.
        m.d.comb += self.stream.payload.eq(mem_read_port.data)
        with m.If(~self.stream.valid | self.stream.ready):
            # The stream produces the next word when there is data in the FIFO.
            m.d.comb += [
                mem_read_port.en    .eq(fifo_word_count != 0)
            ]
            m.d.sync += [
                self.stream.valid   .eq(mem_read_port.en),
                self.stream.last    .eq(fifo_word_count == 1),
            ]
        with m.Else():
            m.d.comb += mem_read_port.en.eq(0)

        # When a word is read from the FIFO, move to the next address.
        m.d.comb += mem_read_port.addr.eq(read_word_addr)
        with m.If(mem_read_port.en):
            m.d.sync += read_word_addr.eq(read_word_addr + 1)

        #
        # FIFO count handling.
        #

        # Number of words popped from the FIFO this cycle.
        fifo_words_popped = Signal(1)

        # Number of uncommitted words and its push trigger.
        fifo_words_pending = Signal(11)
        data_commit  = Signal()

        # One word is popped if the FIFO stream is read.
        m.d.comb += fifo_words_popped.eq(mem_read_port.en)

        # If discarding data, set the count to zero.
        with m.If(self.discarding):
            m.d.usb += [
                write_byte_addr.eq(0),
            ]
            m.d.sync += [
                self.stream.valid.eq(0),
                fifo_word_count.eq(0),
                read_word_addr.eq(0),
                fifo_words_pending.eq(0),
            ]
        # Otherwise, update the count acording to words pushed and popped.
        with m.Else():
            fifo_next_count = fifo_word_count - fifo_words_popped
            with m.If(data_commit):
                m.d.sync += fifo_word_count.eq(fifo_next_count + fifo_words_pending)
            with m.Else():
                m.d.sync += fifo_word_count.eq(fifo_next_count)

        # Timestamp counter.
        current_time = Signal(16)
        m.d.usb += current_time.eq(current_time + 1)

        #
        # Core analysis FSM.
        #
        with m.FSM(domain="usb") as f:
            m.d.comb += [
                self.idle      .eq(f.ongoing("AWAIT_START") | f.ongoing("AWAIT_PACKET")),
                self.stopped   .eq(f.ongoing("AWAIT_START")),
                self.overrun   .eq(f.ongoing("OVERRUN")),
                self.capturing .eq(f.ongoing("CAPTURE_PACKET")),
                self.discarding.eq(self.stopped & self.capture_enable),
            ]

            # AWAIT_START: wait for capture to be enabled, but don't start mid-packet.
            with m.State("AWAIT_START"):
                with m.If(self.capture_enable & ~self.utmi.rx_active):
                    m.next = "AWAIT_PACKET"
                    m.d.usb += current_time.eq(0)


            # AWAIT_PACKET: capture is enabled, wait for a packet to start.
            with m.State("AWAIT_PACKET"):
                with m.If(~self.capture_enable):
                    m.next = "AWAIT_START"
                with m.Elif(self.utmi.rx_active):
                    m.next = "CAPTURE_PACKET"
                    m.d.usb += [
                        header_word_addr   .eq(next_word_addr),
                        write_byte_addr    .eq(write_byte_addr + write_odd + self.HEADER_SIZE_BYTES),
                        packet_size        .eq(0),
                        packet_time        .eq(current_time),
                        current_time       .eq(0),
                    ]
                    m.d.sync += [
                        fifo_words_pending .eq(self.HEADER_SIZE_WORDS),
                    ]
                with m.Elif(current_time == 0xFFFF):
                    # The timestamp is about to wrap. Write a dummy event.
                    m.d.comb += [
                        write_event        .eq(1),
                        event_code         .eq(USBAnalyzerEvent.NONE),
                    ]
                    m.d.usb += [
                        write_byte_addr    .eq(write_byte_addr + write_odd + self.EVENT_SIZE_BYTES),
                    ]
                    m.d.sync += [
                        fifo_words_pending .eq(self.EVENT_SIZE_WORDS),
                    ]


            # Capture data until the packet is complete.
            with m.State("CAPTURE_PACKET"):

                byte_received = self.utmi.rx_valid & self.utmi.rx_active

                # Capture data whenever RxValid is asserted.
                m.d.comb += [
                    write_packet    .eq(byte_received),
                ]

                # Advance the write pointer each time we receive a bit.
                with m.If(byte_received):
                    m.d.usb += [
                        write_byte_addr    .eq(write_byte_addr + 1),
                        packet_size        .eq(packet_size + 1),
                    ]

                    # If this would be filling up our data memory,
                    # move to the OVERRUN state.
                    with m.If(fifo_word_count + fifo_words_pending == self.mem_size_words - 1):
                        m.next = "OVERRUN"

                # If we've stopped receiving, write header.
                with m.If(~self.utmi.rx_active):
                    m.d.comb += [
                        write_header .eq(1),
                    ]
                    m.next = "AWAIT_PACKET"


            # BABBLE -- handles the case in which we've received a packet beyond
            # the allowable size in the USB spec
            with m.State("BABBLE"):

                # Trap here, for now.
                pass


            with m.State("OVERRUN"):
                # TODO: we should probably set an overrun flag and then emit an EOP, here?

                # If capture is stopped by the host, reset back to the ready state.
                with m.If(~self.capture_enable):
                    m.next = "AWAIT_START"


        #
        # Buffer write FSM.
        #
        with m.FSM(domain="sync"):
            # START: Begin write operation when requested.
            with m.State("START"):
                with m.If(write_packet):
                    # Write packet byte.
                    m.d.comb += [
                        mem_write_port.addr  .eq(write_word_addr),
                        mem_write_port.data  .eq(self.utmi.rx_data.replicate(2)),
                        mem_write_port.en    .eq(Mux(write_odd, 0b01, 0b10)),
                    ]
                    m.d.sync += [
                        fifo_words_pending   .eq(fifo_words_pending + ~write_odd),
                    ]
                    m.next = "IDLE"
                with m.Elif(write_header):
                    # Write first word of header.
                    m.d.comb += [
                        mem_write_port.addr  .eq(header_word_addr),
                        mem_write_port.data  .eq(packet_size),
                        mem_write_port.en    .eq(0b11)
                    ]
                    m.next = "FINISH_HEADER"
                with m.Elif(write_event):
                    # Write event identifier and event code.
                    m.d.comb += [
                        mem_write_port.addr  .eq(next_word_addr),
                        mem_write_port.data  .eq(Cat([event_code, C(0xFF, 8)])),
                        mem_write_port.en    .eq(0b11),
                    ]
                    m.next = "FINISH_EVENT"

            # FINISH_HEADER: Write second word of header.
            with m.State("FINISH_HEADER"):
                m.d.comb += [
                        mem_write_port.addr  .eq(header_word_addr + 1),
                        mem_write_port.data  .eq(packet_time),
                        mem_write_port.en    .eq(0b11),
                        data_commit          .eq(1),
                ]
                m.next = "START"

            # FINISH_EVENT: Write second word of event.
            with m.State("FINISH_EVENT"):
                m.d.comb += [
                        mem_write_port.addr  .eq(next_word_addr + 1),
                        mem_write_port.data  .eq(current_time),
                        mem_write_port.en    .eq(0b11),
                        data_commit          .eq(1),
                ]
                m.next = "START"

            # IDLE: Nothing to do this cycle.
            with m.State("IDLE"):
                m.next = "START"


        return m


class USBAnalyzerEvent(IntEnum):
    NONE = 0


class USBAnalyzerTestBase(LunaGatewareTestCase):

    SYNC_CLOCK_FREQUENCY = 120e6
    USB_CLOCK_FREQUENCY = 60e6

    def expect_data(self, expected_data):
        # Check the stream reports data available.
        self.assertEqual((yield self.stream.valid), 1)

        # Check that the expected data is set up.
        self.assertEqual((yield self.stream.payload), expected_data[0])

        # Signal that we are ready to receive data.
        yield self.stream.ready.eq(1)
        yield

        # Validate that we get all of the bytes we expected.
        received_data = []
        for datum in expected_data:
            if (yield self.stream.valid):
                received_data.append((yield self.stream.payload))
                yield
            else:
                # Data ended early.
                break
        self.assertEqual(received_data, expected_data)

        if len(expected_data) % 2 == 1:
            # There should then be one padding byte.
            self.assertEqual((yield self.stream.valid), 1)
            yield

        # There should then be no data left.
        self.assertEqual((yield self.stream.valid), 0)


class USBAnalyzerTest(USBAnalyzerTestBase):

    def instantiate_dut(self):

        from amaranth import DomainRenamer, ResetInserter

        self.utmi = Record([
            ('tx_data',     8),
            ('rx_data',     8),

            ('rx_valid',    1),
            ('rx_active',   1),
            ('rx_error',    1),
            ('rx_complete', 1),
        ])
        m = Module()
        m.submodules.analyzer = self.analyzer = USBAnalyzer(utmi_interface=self.utmi, mem_depth=128)

        reset_on_start = ResetInserter(self.analyzer.discarding)
        m.submodules.s16to8 = s16to8 = reset_on_start(Stream16to8())
        m.submodules.clk_conv = clk_conv = StreamFIFO(
            AsyncFIFOReadReset(width=8, depth=4, r_domain="usb", w_domain="sync"))
        m.d.comb += [
            s16to8.input.stream_eq(self.analyzer.stream),
            clk_conv.input.stream_eq(s16to8.output),
            clk_conv.fifo.ext_rst.eq(self.analyzer.discarding),
        ]
        self.stream = clk_conv.output
        return m


    def advance_stream(self, value):
        yield self.utmi.rx_data.eq(value)
        yield


    @usb_domain_test_case
    def test_single_packet(self):
        # Enable capture
        yield self.analyzer.capture_enable.eq(1)
        yield

        # Ensure we're not capturing until a transaction starts.
        self.assertEqual((yield self.analyzer.capturing), 0)

        # Apply our first input, and validate that we start capturing.
        yield self.utmi.rx_active.eq(1)
        yield self.utmi.rx_valid.eq(1)
        yield self.utmi.rx_data.eq(0)
        yield
        yield

        # Provide some data.
        for i in range(1, 10):
            yield from self.advance_stream(i)
            self.assertEqual((yield self.analyzer.capturing), 1)

        # Ensure we're still capturing, _and_ that we have
        # data available.
        self.assertEqual((yield self.analyzer.capturing), 1)

        # End our packet.
        yield self.utmi.rx_active.eq(0)
        yield from self.advance_stream(10)

        # Idle for several cycles.
        yield from self.advance_cycles(5)
        self.assertEqual((yield self.analyzer.capturing), 0)

        # First, we should get a header with the total data length.
        # This should be 0x00, 0x0a; as we captured 10 bytes.
        #
        # Next, we should get a timestamp with the cycle count at which
        # the packet started. This should be zero.
        #
        # Finally, there should be the 10 packet bytes.
        yield from self.expect_data([0x00, 0x0a, 0x00, 0x00] + list(range(0, 10)))


    @usb_domain_test_case
    def test_slow_packet(self):
        # Enable capture
        yield self.analyzer.capture_enable.eq(1)
        yield

        # Start a packet.
        yield self.utmi.rx_active.eq(1)
        yield

        # Provide a byte every 40 cycles, as in full speed capture.
        for byte in range(0, 10):
            yield from self.advance_cycles(39)
            yield self.utmi.rx_data.eq(byte)
            yield self.utmi.rx_valid.eq(1)
            yield
            yield self.utmi.rx_valid.eq(0)

        # End our packet.
        yield self.utmi.rx_active.eq(0)

        # Idle for several cycles.
        yield from self.advance_cycles(5)

        # First, we should get a header with the total data length.
        # This should be 0x00, 0x0a; as we captured 10 bytes.
        #
        # Next, we should get a timestamp with the cycle count at which
        # the packet started. This should be zero.
        #
        # Finally, there should be the 10 packet bytes.
        yield from self.expect_data([0x00, 0x0a, 0x00, 0x00] + list(range(0, 10)))


    @usb_domain_test_case
    def test_short_packet(self):
        # Enable capture
        yield self.analyzer.capture_enable.eq(1)
        yield

        # Apply our first input, and validate that we start capturing.
        yield self.utmi.rx_active.eq(1)
        yield self.utmi.rx_valid.eq(1)
        yield self.utmi.rx_data.eq(0)
        yield

        # Provide some data.
        yield from self.advance_stream(0xAB)

        # End our packet.
        yield self.utmi.rx_active.eq(0)
        yield from self.advance_stream(10)

        # Idle for several cycles.
        yield from self.advance_cycles(5)
        self.assertEqual((yield self.analyzer.capturing), 0)

        # First, we should get a header with the total data length.
        # This should be 0x00, 0x01; as we captured 1 byte.
        # Next, we should get a timestamp with the cycle count at which
        # the packet started. This should be 0x00, 0x00.
        yield from self.expect_data([0x00, 0x01, 0x00, 0x00, 0xab])


    @usb_domain_test_case
    def test_timestamp_wrap(self):
        # Enable capture.
        yield self.analyzer.capture_enable.eq(1)
        yield

        # Nothing happens for 0x10123 cycles.
        yield from self.advance_cycles(0x10123)

        # Then there's a one-byte packet.
        yield self.utmi.rx_active.eq(1)
        yield self.utmi.rx_valid.eq(1)
        yield self.utmi.rx_data.eq(0)
        yield
        yield from self.advance_stream(0xAB)
        yield self.utmi.rx_active.eq(0)
        yield from self.advance_stream(10)

        # First, we should get an event with code zero, timestamp 0xFFFF.
        # Next we should get the packet, with length 1 and timestamp 0x0123.
        rollover_event = [0xFF, 0x00, 0xFF, 0xFF]
        packet = [0x00, 0x01, 0x01, 0x23, 0xAB]
        yield from self.expect_data(rollover_event + packet)


class USBAnalyzerStackTest(USBAnalyzerTestBase):
    """ Test that evaluates a full-stack USB analyzer setup. """

    def instantiate_dut(self):

        from luna.gateware.interface.ulpi import UTMITranslator
        from amaranth import DomainRenamer, ResetInserter

        self.ulpi = Record([
            ('data', [
                ('i',  8),
                ('o',  8),
                ('oe', 8),
            ]),
            ('nxt', [('i', 1)]),
            ('stp', [('o', 1)]),
            ('dir', [('i', 1)]),
            ('clk', [('o', 1)]),
            ('rst', [('o', 1)]),
        ])

        # Create a stack of our UTMITranslator and our USBAnalyzer.
        # We'll wrap the both in a module to establish a synthetic hierarchy.
        m = Module()
        m.submodules.translator = self.translator = UTMITranslator(ulpi=self.ulpi, handle_clocking=False)
        m.submodules.analyzer   = self.analyzer   = USBAnalyzer(utmi_interface=self.translator, mem_depth=128)
        reset_on_start = ResetInserter(self.analyzer.discarding)
        m.submodules.s16to8 = s16to8 = reset_on_start(Stream16to8())
        m.submodules.clk_conv = clk_conv = StreamFIFO(
            AsyncFIFOReadReset(width=8, depth=4, r_domain="usb", w_domain="sync"))
        m.d.comb += [
            s16to8.input.stream_eq(self.analyzer.stream),
            clk_conv.input.stream_eq(s16to8.output),
            clk_conv.fifo.ext_rst.eq(self.analyzer.discarding),
        ]
        self.stream = clk_conv.output
        return m


    def initialize_signals(self):

        # Ensure the translator doesn't need to perform any register reads/writes
        # by default, so we can focus on packet Rx.
        yield self.translator.xcvr_select.eq(1)
        yield self.translator.dm_pulldown.eq(1)
        yield self.translator.dp_pulldown.eq(1)
        yield self.translator.use_external_vbus_indicator.eq(0)


    @usb_domain_test_case
    def test_simple_analysis(self):
        # Enable capture
        yield self.analyzer.capture_enable.eq(1)
        yield from self.advance_cycles(10)

        # Start a new packet.
        yield self.ulpi.dir.i.eq(1)
        yield self.ulpi.nxt.i.eq(1)

        # Bus turnaround packet.
        yield self.ulpi.data.i.eq(0x80)
        yield

        # Provide some data to be captured.
        for i in [0x2d, 0x00, 0x10]:
            yield self.ulpi.data.i.eq(i)
            yield

        # Mark our packet as complete.
        yield self.ulpi.dir.i.eq(0)
        yield self.ulpi.nxt.i.eq(0)
        yield

        # Wait for a few cycles, for realism.
        yield from self.advance_cycles(10)

        # Validate that we got the correct packet out; plus headers.
        # We waited 10 cycles before starting the packet, so the
        # timestamp should be 0x00, 0x0a.
        yield from self.expect_data([0x00, 0x03, 0x00, 0x0a, 0x2d, 0x00, 0x10])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
#
# This file is part of Cynthion.
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Cynthion 'update' command. """

from .util import find_cynthion_asset, find_cynthion_bitstream
from .util import flash_bitstream, flash_soc_firmware, flash_mcu_firmware


def cynthion_update(device, args):
    if args.bitstream:
        flash_bitstream(device, find_cynthion_bitstream(device, "analyzer.bit"))
    elif args.mcu_firmware:
        flash_mcu_firmware(device, find_cynthion_asset("apollo.bin"))
    else:
        flash_bitstream(device, find_cynthion_bitstream(device, "analyzer.bit"))
        flash_mcu_firmware(device, find_cynthion_asset("apollo.bin"))



#!/usr/bin/env python
#
# This file is part of Cynthion.
#
# Copyright (c) 2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Cynthion CLI commands. """

from .cynthion_build           import *
from .cynthion_info            import *
from .cynthion_flash           import *
from .cynthion_run             import *
from .cynthion_setup           import *
from .cynthion_update          import *



#
# This file is part of Cynthion
#

import os

from ..board import CynthionBoard

class CynthionMoondancer(CynthionBoard):
    """ Class representing Cynthion base-boards running the Moondancer firmware. """

    # This field refers to the USB device version number
    #
    # digit 1:   major
    # digit 2:   minor
    # digit 3&4: patch version
    HANDLED_BOARD_VERSIONS = [
        0x0004, # r0.4
        0x0005, # r0.5
        0x0006, # r0.6
        0x0007, # r0.7
        0x1000, # r1.0.0
        0x1100, # r1.1.0
        0x1101, # r1.1.1
    ]

    # Currently, all Cynthion Moondancer boards have an ID of 0x10.
    HANDLED_BOARD_IDS = [0x10]

    BOARD_NAME = "Cynthion in Moondancer mode"

    # The Cynthion has six LEDs.
    SUPPORTED_LEDS = 6

    # All of the GPIO mappings accessible from the Cynthion pmod headers.
    # TODO: add pmod gpio mappings for cynthion
    GPIO_MAPPINGS = {
    }

    def initialize_apis(self):
        """ Initialize a new Cynthion connection. """

        # Set up the core connection.
        super(CynthionMoondancer, self).initialize_apis()

        # Create our simple peripherals.
        self._populate_simple_interfaces()

        # Initialize the fixed peripherals that come on the board...

        # Populate the per-board GPIO.
        if self.supports_api("gpio"):
            self._populate_gpio()

        # Add objects for each of our LEDs.
        self._populate_leds(self.SUPPORTED_LEDS)



#
# This file is part of Cynthion.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from amaranth                       import Elaboratable, Module, Signal, Cat, Memory, ResetSignal
from amaranth.lib.fifo              import SyncFIFOBuffered, FIFOInterface
from amaranth.lib.coding            import GrayDecoder, GrayEncoder
from amaranth.hdl.ast               import Assume, Initial
from amaranth.lib.cdc               import FFSynchronizer
from amaranth.utils                 import log2_int

from luna.gateware.stream           import StreamInterface
from luna.gateware.interface.psram  import HyperRAMInterface, HyperRAMPHY


class StreamFIFO(Elaboratable):
    def __init__(self, fifo):
        self.fifo   = fifo
        self.input  = StreamInterface(payload_width=fifo.width)
        self.output = StreamInterface(payload_width=fifo.width)

    def elaborate(self, platform):
        m = Module()

        m.submodules.fifo = self.fifo

        m.d.comb += [
            self.fifo.w_data    .eq(self.input.payload),
            self.fifo.w_en      .eq(self.input.valid),
            self.input.ready    .eq(self.fifo.w_rdy),

            self.output.payload .eq(self.fifo.r_data),
            self.output.valid   .eq(self.fifo.r_rdy),
            self.fifo.r_en      .eq(self.output.ready),
        ]

        return m


class HyperRAMPacketFIFO(Elaboratable):
    def __init__(self, out_fifo_depth=None):
        self.input  = StreamInterface(payload_width=16)
        self.output = StreamInterface(payload_width=16)
        # A minimum output FIFO depth of 2 prevents data loss during consumer stalls.
        self.out_fifo_depth = max(out_fifo_depth, 2) if out_fifo_depth is not None else 2

    def elaborate(self, platform):
        m = Module()

        # HyperRAM submodules
        ram_bus         = platform.request('ram')
        psram_phy       = HyperRAMPHY(bus=ram_bus)
        psram           = HyperRAMInterface(phy=psram_phy.phy)
        m.submodules   += [psram_phy, psram]

        # HyperRAM status
        depth         = 2 ** 22
        write_address = Signal(range(depth))
        read_address  = Signal(range(depth))
        word_count    = Signal(range(depth + 1))
        empty         = Signal()
        full          = Signal()
        m.d.comb += [
            empty .eq(word_count == 0),
            full  .eq(word_count == depth),
        ]

        # Update word count and pointers using the write and read strobes.
        m.d.sync += word_count.eq(word_count - psram.read_ready + psram.write_ready)
        with m.If(psram.read_ready):
            m.d.sync += read_address.eq(read_address + 1)
        with m.If(psram.write_ready):
            m.d.sync += write_address.eq(write_address + 1)

        # This output buffer prevents data loss during consumer stalls. It can also be used
        # to gather entire bursts from the HyperRAM if `out_fifo_depth` is big enough.
        m.submodules.out_fifo = out_fifo = SyncFIFOBuffered(width=16, depth=self.out_fifo_depth)

        # Hook up our PSRAM.
        m.d.comb += [
            ram_bus.reset.o       .eq(0),
            psram.single_page     .eq(0),
            psram.register_space  .eq(0),
            psram.write_data      .eq(self.input.payload),
            self.input.ready      .eq(psram.write_ready),

            # Wire PSRAM -> output FIFO -> output stream
            out_fifo.w_data       .eq(psram.read_data),
            out_fifo.w_en         .eq(psram.read_ready),
            self.output.payload   .eq(out_fifo.r_data),
            self.output.valid     .eq(out_fifo.r_rdy),
            out_fifo.r_en         .eq(self.output.ready),
        ]

        # Generation of the final word condition.
        is_write = Signal()
        with m.If(is_write):
            # WRITE: Finish when there's no space or incoming data.
            m.d.comb += psram.final_word.eq((word_count == (depth-1)) | self.input.last)
        with m.Else():
            # READ: Finish when PSRAM is empty or the output FIFO is full.
            m.d.comb += psram.final_word.eq((word_count == 1) | (out_fifo.level == out_fifo.depth - 1))

        #
        # HyperRAM Packet FIFO state machine
        #
        with m.FSM(domain="sync"):

            # IDLE: Begin a write / read burst operation when ready.
            with m.State("IDLE"):
                # Write whenever we have input data...
                with m.If(self.input.valid & ~full):
                    m.d.comb += [
                        psram.address           .eq(write_address),
                        psram.perform_write     .eq(1),
                        psram.start_transfer    .eq(1),
                    ]
                    m.d.sync += is_write.eq(1)
                    m.next = "BUSY"

                # ...otherwise, read when FIFO is less than half full.
                with m.Elif(~empty & (out_fifo.level[-1] == 0)):
                    m.d.comb += [
                        psram.address           .eq(read_address),
                        psram.perform_write     .eq(0),
                        psram.start_transfer    .eq(1),
                    ]
                    m.d.sync += is_write.eq(0)
                    m.next = "BUSY"

            # BUSY: Wait for the PSRAM to recover before a new transaction.
            with m.State("BUSY"):
                with m.If(psram.idle):
                    m.next = "IDLE"

        return m


class Stream16to8(Elaboratable):
    def __init__(self, msb_first=True):
        self.msb_first = msb_first
        self.input     = StreamInterface(payload_width=16)
        self.output    = StreamInterface(payload_width=8)

    def elaborate(self, platform):
        m = Module()

        input_data = self.input.payload
        if self.msb_first:
            input_data = Cat(input_data[8:16], input_data[0:8])

        odd_byte   = Signal()
        data_shift = Signal.like(self.input.payload)  # shift register
        m.d.comb  += self.output.payload.eq(data_shift[0:8])

        # When the output stream is not stalled...
        with m.If(self.output.ready | ~self.output.valid):

            # If odd_byte is asserted, send the buffered second byte
            with m.If(odd_byte):
                m.d.sync += [
                    data_shift          .eq(data_shift[8:]),
                    self.output.valid   .eq(1),
                    odd_byte            .eq(0),
                ]

            # Otherwise, consume a new word from the input stream
            with m.Else():
                m.d.comb += self.input.ready .eq(1)
                m.d.sync += self.output.valid.eq(self.input.valid)
                with m.If(self.input.valid):
                    m.d.sync += [
                        data_shift .eq(input_data),
                        odd_byte   .eq(1),
                    ]

        return m


class AsyncFIFOReadReset(Elaboratable, FIFOInterface):
    def __init__(self, *, width, depth, r_domain="read", w_domain="write", exact_depth=False):
        if depth != 0:
            try:
                depth_bits = log2_int(depth, need_pow2=exact_depth)
                depth = 1 << depth_bits
            except ValueError:
                raise ValueError("AsyncFIFO only supports depths that are powers of 2; requested "
                                 "exact depth {} is not"
                                 .format(depth)) from None
        else:
            depth_bits = 0
        super().__init__(width=width, depth=depth)

        self.ext_rst = Signal()
        self._r_domain = r_domain
        self._w_domain = w_domain
        self._ctr_bits = depth_bits + 1

    def elaborate(self, platform):
        m = Module()
        if self.depth == 0:
            m.d.comb += [
                self.w_rdy.eq(0),
                self.r_rdy.eq(0),
            ]
            return m

        # The design of this queue is the "style #2" from Clifford E. Cummings' paper "Simulation
        # and Synthesis Techniques for Asynchronous FIFO Design":
        # http://www.sunburst-design.com/papers/CummingsSNUG2002SJ_FIFO1.pdf

        do_write = self.w_rdy & self.w_en
        do_read  = self.r_rdy & self.r_en

        # TODO: extract this pattern into lib.cdc.GrayCounter
        produce_w_bin = Signal(self._ctr_bits)
        produce_w_nxt = Signal(self._ctr_bits)
        m.d.comb += produce_w_nxt.eq(produce_w_bin + do_write)
        m.d[self._w_domain] += produce_w_bin.eq(produce_w_nxt)

        # Note: Both read-domain counters must be reset_less (see comments below)
        consume_r_bin = Signal(self._ctr_bits, reset_less=True)
        consume_r_nxt = Signal(self._ctr_bits)
        m.d.comb += consume_r_nxt.eq(consume_r_bin + do_read)
        m.d[self._r_domain] += consume_r_bin.eq(consume_r_nxt)

        produce_w_gry = Signal(self._ctr_bits)
        produce_r_gry = Signal(self._ctr_bits)
        produce_enc = m.submodules.produce_enc = \
            GrayEncoder(self._ctr_bits)
        produce_cdc = m.submodules.produce_cdc = \
            FFSynchronizer(produce_w_gry, produce_r_gry, o_domain=self._r_domain)
        m.d.comb += produce_enc.i.eq(produce_w_nxt),
        m.d[self._w_domain] += produce_w_gry.eq(produce_enc.o)

        consume_r_gry = Signal(self._ctr_bits, reset_less=True)
        consume_w_gry = Signal(self._ctr_bits)
        consume_enc = m.submodules.consume_enc = \
            GrayEncoder(self._ctr_bits)
        consume_cdc = m.submodules.consume_cdc = \
            FFSynchronizer(consume_r_gry, consume_w_gry, o_domain=self._w_domain)
        m.d.comb += consume_enc.i.eq(consume_r_nxt)
        m.d[self._r_domain] += consume_r_gry.eq(consume_enc.o)

        consume_w_bin = Signal(self._ctr_bits)
        consume_dec = m.submodules.consume_dec = \
            GrayDecoder(self._ctr_bits)
        m.d.comb += consume_dec.i.eq(consume_w_gry),
        m.d[self._w_domain] += consume_w_bin.eq(consume_dec.o)

        produce_r_bin = Signal(self._ctr_bits)
        produce_dec = m.submodules.produce_dec = \
            GrayDecoder(self._ctr_bits)
        m.d.comb += produce_dec.i.eq(produce_r_gry),
        m.d.comb += produce_r_bin.eq(produce_dec.o)

        w_full  = Signal()
        r_empty = Signal()
        m.d.comb += [
            w_full.eq((produce_w_gry[-1]  != consume_w_gry[-1]) &
                      (produce_w_gry[-2]  != consume_w_gry[-2]) &
                      (produce_w_gry[:-2] == consume_w_gry[:-2])),
            r_empty.eq(consume_r_gry == produce_r_gry),
        ]

        m.d[self._w_domain] += self.w_level.eq(produce_w_bin - consume_w_bin)
        m.d.comb += self.r_level.eq(produce_r_bin - consume_r_bin)

        storage = Memory(width=self.width, depth=self.depth)
        w_port  = m.submodules.w_port = storage.write_port(domain=self._w_domain)
        r_port  = m.submodules.r_port = storage.read_port (domain=self._r_domain,
                                                           transparent=False)
        m.d.comb += [
            w_port.addr.eq(produce_w_bin[:-1]),
            w_port.data.eq(self.w_data),
            w_port.en.eq(do_write),
            self.w_rdy.eq(~w_full),
        ]
        m.d.comb += [
            r_port.addr.eq(consume_r_nxt[:-1]),
            self.r_data.eq(r_port.data),
            r_port.en.eq(1),
            self.r_rdy.eq(~r_empty),
        ]

        # Reset handling differs from Amaranth's AsyncFIFO: reset control rests entirely with 
        # the read domain. An additional synchronous external reset signal is also included.
        r_rst = ResetSignal(domain=self._r_domain, allow_reset_less=True)

        # Decode Gray code counter to overwrite binary counter in read domain.
        rst_dec = m.submodules.rst_dec = \
            GrayDecoder(self._ctr_bits)
        m.d.comb += rst_dec.i.eq(produce_r_gry)
        with m.If(r_rst | self.ext_rst):
            m.d.comb += r_empty.eq(1)
            m.d[self._r_domain] += consume_r_gry.eq(produce_r_gry)
            m.d[self._r_domain] += consume_r_bin.eq(rst_dec.o)

        if platform == "formal":
            with m.If(Initial()):
                m.d.comb += Assume(produce_w_gry == (produce_w_bin ^ produce_w_bin[1:]))
                m.d.comb += Assume(consume_r_gry == (consume_r_bin ^ consume_r_bin[1:]))

        return m



#
# This file is part of Cynthion
#

"""
Module containing the core definitions for a Cynthion board.
"""

import string
from weakref import WeakSet

from pygreat.board import GreatBoard

from .interfaces.led import LED
from .interfaces.gpio import GPIO

from .shared import usb

# Default device identifiers.
CYNTHION_VENDOR_ID  = usb.bVendorId.cynthion
CYNTHION_PRODUCT_ID = usb.bProductId.cynthion

# libgreat backend interface subclass
LIBGREAT_INTERFACE_SUBCLASS = usb.bInterfaceSubClass.moondancer

# libgreat backend interface protocol version
LIBGREAT_INTERFACE_PROTOCOL = usb.bInterfaceProtocol.moondancer

# Quirk constant that helps us identify libusb's pipe errors, which bubble
# up as generic USBErrors with errno 32 on affected platforms.
LIBUSB_PIPE_ERROR = 32

# Total seconds we should wait after a reset before reconnecting.
RECONNECT_DELAY = 3


class CynthionBoard(GreatBoard):
    """
    Class describing Cynthion devices.
    """

    # Default device identifiers.
    BOARD_VENDOR_ID = CYNTHION_VENDOR_ID
    BOARD_PRODUCT_ID = CYNTHION_PRODUCT_ID

    """
    The mappings from GPIO names to port numbers. Paths in names can be delineated
    with underscores to group gpios. For example, if Jumper 7, Pin 3 is Port 5, Pin 11,
    you could add an entry that reads "J7_P3": (5, 11).
    """
    GPIO_MAPPINGS = {}

    # FIXME: should these interfaces be in libgreat?

    #
    # Quick way to add simple Python wrappers around comms classes.
    # Create an entry for the relevant comms class, and provide a tuple of
    #   (attr_name, peripheral_class).
    #
    SIMPLE_CLASS_MAPPINGS = {
        # TODO 'firmware': ('onboard_flash', DeviceFirmwareManager),
        # TODO 'gpio': ('gpio', GPIO),
    }

    def __init__(self, *args, **device_identifiers):
        """ Initialize a new CynthionBoard instance with our additional properties. """

        # Create a new list of interfaces and programmers.
        self._interfaces = []
        self._instantiated_programmers = WeakSet()

        # Add libgreat usb interface protocol to device identifiers.
        device_identifiers['interface_subclass'] = LIBGREAT_INTERFACE_SUBCLASS
        device_identifiers['interface_protocol'] = LIBGREAT_INTERFACE_PROTOCOL

        super(CynthionBoard, self).__init__(*args, **device_identifiers)


    def available_interfaces(self):
        """ Returns a list of peripheral properties that exist on this board. """
        return self._interfaces[:]


    def _populate_leds(self, led_count):
        """Adds the standard set of LEDs to the board object.

        Args:
            led_count -- The number of LEDS present on the board.
        """

        self._add_interface('leds', {})

        for i in range(1, led_count + 1):
            self.leds[i] = LED(self, i)


    def _populate_gpio(self):
        """ Adds GPIO pin definitions to the board's main GPIO object. """

        # Handle each GPIO mapping.
        for name, pin in self.GPIO_MAPPINGS.items():
            self.gpio.register_gpio(name, pin)


    def _add_interface(self, name, instance):
        """
        Adds a peripheral to the Cynthion object. Prefer this over adding attributes directly,
        as it adds peripherals to a list that can be queried by the user.

        Arguments:
            name -- The name of the attribute to add to this board. "i2c" would create a
                .i2c property on this board.
            instance -- The object to add as that property.
        """

        self._interfaces.append(name)
        setattr(self, name, instance)


    def _add_simple_interface(self, name, cls, *args, **kwargs):
        """ Adds a given interface to this board.

        Arguments:
            name -- The attribute name to be added to the board.
            cls -- The class to be instantiated to create the given object.
        """

        # Create an instance of the relevant peripheral class...
        instance = cls(self, *args, **kwargs)

        # ... and add it to this board.
        self._add_interface(name, instance)



    def _populate_simple_interfaces(self):
        """ Adds simple interfaces to the board object by parsing the SIMPLE_CLASS_MAPPINGS dictionary. """

        for comms_class, interface in self.SIMPLE_CLASS_MAPPINGS.items():

            # If the relevant API is supported, add the relevant peripheral.
            if self.supports_api(comms_class):
                name, python_class = interface
                self._add_simple_interface(name, python_class)

    def available_programmers(self, as_dictionary=False):
        """ Returns the list of available programmers. """

        from types import ModuleType

        programmers = {}


        for module in ProgrammerModules.__dict__.values():
            if isinstance(module, ModuleType) and hasattr(module, 'create_programmer'):
                module_name = module.__name__.split('.')[-1]
                programmers[module_name] = module

        if as_dictionary:
            return programmers
        else:
            return list(programmers.values())


    def create_programmer(self, name, *args, **kwargs):
        """ Creates a new instance of the programmer with the given name. """

        try:
            programmer_module = self.available_programmers(True)[name]
            programmer = programmer_module.create_programmer(self, *args, **kwargs)

            # Keep a weak reference to the relevant programmer.
            # This is useful for re-attaching programmers after a disconnect.
            self._instantiated_programmers.add(programmer)

            # Finally, return the created programmer.
            return programmer

        except KeyError:
            raise KeyError("no available programmer named {}".format(name))


    def __dir__(self):
        """ Generate a cleaned-up dir listing for the relevant board. """

        items = super(CynthionBoard, self).__dir__()
        return [item for item in items if item[0] in string.ascii_lowercase]
